# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from collections import namedtuple
import json
import logging
import flask
from superdesk import get_resource_service
import superdesk
from superdesk.errors import SuperdeskApiError
from superdesk.metadata.item import not_analyzed, ITEM_STATE, PUBLISH_STATES
from superdesk.metadata.utils import aggregations, get_elastic_highlight_query
from superdesk.resource import Resource
from superdesk.services import BaseService
from superdesk.utc import utcnow

from bson.objectid import ObjectId
from eve.utils import ParsedRequest, config
from flask import current_app as app

from apps.archive.archive import SOURCE as ARCHIVE
from apps.archive.common import handle_existing_data, item_schema
from superdesk.publish.publish_queue import PUBLISHED_IN_PACKAGE


logger = logging.getLogger(__name__)

PUBLISHED = 'published'
LAST_PUBLISHED_VERSION = 'last_published_version'
QUEUE_STATE = 'queue_state'
ERROR_MESSAGE = 'error_message'
queue_states = ['pending', 'in_progress', 'queued', 'queued_not_transmitted', 'error']
PUBLISH_STATE = namedtuple('PUBLISH_STATE', ['PENDING',
                                             'IN_PROGRESS',
                                             'QUEUED',
                                             'QUEUED_NOT_TRANSMITTED',
                                             'ERROR'])(*queue_states)

published_item_fields = {
    'item_id': {
        'type': 'string',
        'mapping': not_analyzed
    },
    'publish_state': {
        'type': 'string'
    },
    # last_published_version field is set to true for last published version of the item in the published collection
    # and for the older version is set to false. This field is used to display the last version of the digital copy
    # in the published view.
    LAST_PUBLISHED_VERSION: {
        'type': 'boolean',
        'default': True
    },
    QUEUE_STATE: {
        'type': 'string',
        'default': 'pending',
        'allowed': queue_states,
    },
    ERROR_MESSAGE: {
        'type': 'string'
    },
    'is_take_item': {  # deprecated
        'type': 'boolean',
        'default': False,
    },
    'digital_item_id': {  # deprecated
        'type': 'string'
    },
    'publish_sequence_no': {
        'type': 'integer',
        'readonly': True
    },
    'last_queue_event': {
        'type': 'datetime'
    },
    # item is moved to legal. True if moved else false.
    'moved_to_legal': {
        'type': 'boolean',
        'default': False
    },
    PUBLISHED_IN_PACKAGE: {
        'type': 'string'
    }
}


def get_content_filter():
    """Filter out content of stages not visible to current user (if any).

    :return:
    """
    user = getattr(flask.g, 'user', None)
    if user:
        if 'invisible_stages' in user:
            stages = user.get('invisible_stages')
        else:
            stages = get_resource_service('users').get_invisible_stages_ids(user.get('_id'))

        if stages:
            return {'bool': {'must_not': {'terms': {'task.stage': stages}}}}


class PublishedItemResource(Resource):
    datasource = {
        'search_backend': 'elastic',
        'aggregations': aggregations,
        'es_highlight': get_elastic_highlight_query,
        'default_sort': [('_updated', -1)],
        'elastic_filter_callback': get_content_filter,
        'projection': {
            'old_version': 0,
            'last_version': 0
        }
    }

    schema = item_schema(published_item_fields)
    etag_ignore_fields = [config.ID_FIELD, 'highlights', 'item_id', LAST_PUBLISHED_VERSION, 'moved_to_legal']

    privileges = {'POST': 'publish_queue', 'PATCH': 'publish_queue'}
    item_methods = ['GET', 'PATCH']
    additional_lookup = {
        'url': 'regex("[\w,.:-]+")',
        'field': 'item_id'
    }


class PublishedItemService(BaseService):
    """
    PublishedItemService class is the base class for ArchivedService.
    """

    SEQ_KEY_NAME = 'published_item_sequence_no'

    def on_fetched(self, docs):
        """
        Overriding this to enhance the published article with the one in archive collection
        """

        self.enhance_with_archive_items(docs[config.ITEMS])

    def on_fetched_item(self, doc):
        """
        Overriding this to enhance the published article with the one in archive collection
        """

        self.enhance_with_archive_items([doc])

    def on_create(self, docs):
        """Runs on create.

        An article can be published multiple times in its lifetime. So, it's necessary to preserve the _id which comes
        from archive collection. Also, sets the expiry on the published item and removes the lock information.
        """

        for doc in docs:
            self.raise_if_not_marked_for_publication(doc)
            doc[config.LAST_UPDATED] = doc[config.DATE_CREATED] = utcnow()
            self.set_defaults(doc)

    def on_update(self, updates, original):
        if ITEM_STATE in updates:
            self.raise_if_not_marked_for_publication(updates)

    def raise_if_not_marked_for_publication(self, doc):
        """
        Item should be one of the PUBLISH_STATES. If not raise error.
        """
        if doc.get(ITEM_STATE) not in PUBLISH_STATES:
            raise SuperdeskApiError.badRequestError('Invalid state ({}) for the Published item.'
                                                    .format(doc.get(ITEM_STATE)))

    def set_defaults(self, doc):
        doc['item_id'] = doc[config.ID_FIELD]
        doc['versioncreated'] = utcnow()
        doc['publish_sequence_no'] = get_resource_service('sequences').get_next_sequence_number(self.SEQ_KEY_NAME)
        doc.pop(config.ID_FIELD, None)
        doc.pop('lock_user', None)
        doc.pop('lock_time', None)
        doc.pop('lock_action', None)
        doc.pop('lock_session', None)

    def enhance_with_archive_items(self, items):
        if items:
            ids = list(set([item.get('item_id') for item in items if item.get('item_id')]))
            archive_items = []
            archive_lookup = {}
            if ids:
                query = {'$and': [{config.ID_FIELD: {'$in': ids}}]}
                archive_req = ParsedRequest()
                archive_req.max_results = len(ids)
                # can't access published from elastic due filter on the archive resource hence going to mongo
                archive_items = list(superdesk.get_resource_service(ARCHIVE)
                                     .get_from_mongo(req=archive_req, lookup=query))

                for item in archive_items:
                    handle_existing_data(item)
                    archive_lookup[item[config.ID_FIELD]] = item

            for item in items:
                archive_item = archive_lookup.get(item.get('item_id'), {config.VERSION: item.get(config.VERSION, 1)})

                updates = {
                    config.ID_FIELD: item.get('item_id'),
                    'item_id': item.get(config.ID_FIELD),
                    'lock_user': archive_item.get('lock_user', None),
                    'lock_time': archive_item.get('lock_time', None),
                    'lock_action': archive_item.get('lock_action', None),
                    'lock_session': archive_item.get('lock_session', None),
                    'archive_item': archive_item if archive_item else None
                }

                item.update(updates)
                handle_existing_data(item)

    def on_delete(self, doc):
        """Deleting a published item has a workflow which is implemented in remove_expired().

        Overriding to avoid other services from invoking this method accidentally.
        """

        if app.testing:
            super().on_delete(doc)
        else:
            raise NotImplementedError("Deleting a published item has a workflow which is "
                                      "implemented in remove_expired().")

    def delete_action(self, lookup=None):
        """Deleting a published item has a workflow which is implemented in remove_expired().

        Overriding to avoid other services from invoking this method accidentally.
        """

        if app.testing:
            super().delete_action(lookup)
        else:
            raise NotImplementedError("Deleting a published item has a workflow which is "
                                      "implemented in remove_expired().")

    def on_deleted(self, doc):
        """Deleting a published item has a workflow which is implemented in remove_expired().

        Overriding to avoid other services from invoking this method accidentally.
        """

        if app.testing:
            super().on_deleted(doc)
        else:
            raise NotImplementedError("Deleting a published item has a workflow which is "
                                      "implemented in remove_expired().")

    def get_other_published_items(self, _id):
        try:
            query = {'query': {'filtered': {'filter': {'term': {'item_id': _id}}}}}
            request = ParsedRequest()
            request.args = {'source': json.dumps(query)}
            return super().get(req=request, lookup=None)
        except Exception:
            return []

    def get_last_published_version(self, _id):
        """Returns the last published entry for the passed item id

        :param _id:
        :return:
        """
        try:
            query = {'query':
                     {'filtered':
                      {'filter':
                       {'bool':
                        {'must': [
                            {'term': {'item_id': _id}},
                            {'term': {LAST_PUBLISHED_VERSION: True}}
                        ]}}}}}

            request = ParsedRequest()
            request.args = {'source': json.dumps(query), 'repo': 'published'}
            return get_resource_service('search').get(req=request, lookup=None)
        except Exception:
            return None

    def get_rewritten_items_by_event_story(self, event_id, rewrite_id, rewrite_field):
        """Returns all the rewritten stories from published and archive for a given event and rewrite_id.

        :param str event_id: event id of the document
        :param str rewrite_id: rewrite id of the document
        :param str rewrite_field: field name 'rewrite_of' or 'rewritten_by'
        """
        try:
            query = {'query':
                     {'filtered':
                      {'filter':
                       {'bool':
                        {'must': [
                            {'term': {'event_id': event_id}},
                            {'term': {rewrite_field: rewrite_id}}
                        ]}}}}}

            request = ParsedRequest()
            request.args = {'source': json.dumps(query), 'repo': 'archive,published'}
            return list(get_resource_service('search').get(req=request, lookup=None))
        except Exception:
            return []

    def is_rewritten_before(self, item_id):
        """Checks if the published item is rewritten before.

        :param _id: item_id of the published item
        :return: True is it is rewritten before
        """
        doc = self.find_one(req=None, item_id=item_id)
        return doc and 'rewritten_by' in doc and doc['rewritten_by']

    def update_published_items(self, _id, field, state):
        items = self.get_other_published_items(_id)
        for item in items:
            try:
                super().system_update(ObjectId(item[config.ID_FIELD]), {field: state}, item)
            except Exception:
                # This part is used in unit testing
                super().system_update(item[config.ID_FIELD], {field: state}, item)

    def delete_by_article_id(self, _id):
        """Removes the article from the published collection.

        Removes published queue entries.

        :param str _id: id of the document to be deleted. In mongo, it is the item_id
        """
        lookup = {'item_id': _id}
        self.delete(lookup=lookup)
        get_resource_service('publish_queue').delete_by_article_id(_id)

    def find_one(self, req, **lookup):
        item = super().find_one(req, **lookup)
        handle_existing_data(item)

        return item

    def move_to_archived(self, _id):
        published_items = list(self.get_from_mongo(req=None, lookup={'item_id': _id}))
        if not published_items:
            return
        get_resource_service('archived').post(published_items)
        self.delete_by_article_id(_id)

    def set_moved_to_legal(self, item_id, version, status):
        """Update the legal flag.

        :param str item_id: id of the document
        :param int version: version of the document
        :param boolean status: True if the item is moved to legal else false
        """
        items = self.get_other_published_items(item_id)

        for item in items:
            try:
                if item.get(config.VERSION) <= version and not item.get('moved_to_legal', False):
                    super().system_update(ObjectId(item.get(config.ID_FIELD)), {'moved_to_legal': status}, item)
            except Exception:
                logger.exception('Failed to set the moved_to_legal flag '
                                 'for item {} and version {}'.format(item_id, version))

    def get_published_items_by_moved_to_legal(self, item_ids, move_to_legal):
        """Get the pulished items where flag is moved.

        :param list item_ids: List of item
        :param bool move_to_legal: move_to_legal boolean flag
        :return: list of published items
        """
        if item_ids:
            try:
                query = {
                    'query': {
                        'filtered': {
                            'filter': {
                                'and': [
                                    {'terms': {'item_id': item_ids}},
                                    {'term': {'moved_to_legal': move_to_legal}}
                                ]
                            }
                        }
                    }
                }

                request = ParsedRequest()
                request.args = {'source': json.dumps(query)}
                return list(super().get(req=request, lookup=None))
            except Exception:
                logger.exception('Failed to get published items '
                                 'by moved to legal: {} -- ids: {}.'.format(move_to_legal, item_ids))

        return []




class SearchProvider():
    """Base class for Search Providers.

    You can use ``self.provider`` to get config data.
    """

    #: Provider Type Label
    label = 'unknown'

    def __init__(self, provider):
        """Create search provider instance.

        :param provider: provider data dict
        """
        self.provider = provider

    def find(self, query):
        """Find items using query.

        This must return cursor like object.

        :param query: query dict
        """
        raise NotImplementedError

    def fetch(self, guid):
        """Get single item.

        Get an item before archiving it. Should contain all metadata.

        :param guid: item guid
        """
        raise NotImplementedError

    def fetch_file(self, href):
        """Fetch binary using given href.

        Href is from renditions dict.

        :param href: binary href
        """
        raise NotImplementedError



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013 - 2018 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


from superdesk.io.feed_parsers.image_iptc import ImageIPTCFeedParser
from superdesk.tests import TestCase
import os


class BaseImageIPTCTestCase(TestCase):
    def setUp(self):
        dirname = os.path.dirname(os.path.realpath(__file__))
        self.image_path = os.path.normpath(os.path.join(dirname, '../fixtures', self.filename))
        provider = {'name': 'Test'}
        parser = ImageIPTCFeedParser()
        self.item = parser.parse(self.image_path, provider)


class ImageIPTCTestCase(BaseImageIPTCTestCase):
    filename = 'IPTC-PhotometadataRef-Std2017.1.jpg'

    def test_can_parse(self):
        self.assertTrue(ImageIPTCFeedParser().can_parse(self.image_path))

    def test_content(self):
        item = self.item
        self.assertEqual(item['headline'], 'The Headline (ref2017.1)')
        self.assertEqual(item['byline'], 'Creator1 (ref2017.1)')
        self.assertEqual(item['slugline'], 'The Title (ref2017.1)')
        self.assertEqual(item['description_text'], 'The description aka caption (ref2017.1)',)
        self.assertEqual(item['keywords'], ['Keyword1ref2017.1', 'Keyword2ref2017.1', 'Keyword3ref2017.1'])
        self.assertEqual(item['ednote'], 'An Instruction (ref2017.1)')
        self.assertEqual(item['copyrightnotice'], 'Copyright (Notice) 2017.1 IPTC - www.iptc.org  (ref2017.1)')
        self.assertEqual(item['assignment_id'], 'Job Id (ref2017.1)')
        self.assertEqual(item['firstcreated'].isoformat(), '2017-07-13T17:01:00+00:00')



from superdesk.resource import Resource


class ExportResource(Resource):
    """Export schema"""

    endpoint_name = 'export'
    resource_methods = ['POST']
    item_methods = []
    resource_title = endpoint_name
    schema = {
        'item_ids': {
            'type': 'list',
            'required': True
        },
        'format_type': {
            'type': 'string',
            'required': True
        },
        'validate': {
            'type': 'boolean',
            'required': False
        }
    }
    privileges = {'POST': 'archive'}



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2018 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from copy import deepcopy
from apps.tasks import send_to
from superdesk import register_resource, item_published, get_resource_service, privilege
from superdesk.services import Service
from superdesk.resource import Resource
from superdesk.errors import StopDuplication
from superdesk.metadata.item import PUBLISH_SCHEDULE, SCHEDULE_SETTINGS


NAME = 'internal_destinations'


class InternalDestinationsResource(Resource):
    schema = {
        'name': {'type': 'string', 'required': True},
        'is_active': {'type': 'boolean', 'required': True},
        'filter': Resource.rel('content_filters', nullable=True),
        'desk': Resource.rel('desks', nullable=False, required=True),
        'stage': Resource.rel('stages', nullable=True),
        'macro': {'type': 'string', 'nullable': True},
    }

    privileges = {'POST': 'internal_destinations',
                  'PATCH': 'internal_destinations',
                  'DELETE': 'internal_destinations'}


class InternalDestinationsService(Service):
    pass


def handle_item_published(sender, item, **extra):
    macros_service = get_resource_service('macros')
    archive_service = get_resource_service('archive')
    filters_service = get_resource_service('content_filters')
    destinations_service = get_resource_service(NAME)

    for dest in destinations_service.get(req=None, lookup={'is_active': True}):
        if dest.get('desk') == item.get('task').get('desk'):
            # item desk and internal destination are same then don't execute
            continue

        if dest.get('filter'):
            content_filter = filters_service.find_one(req=None, _id=dest['filter'])
            if not content_filter:  # error state sort of, not sure what to do
                continue
            if not filters_service.does_match(content_filter, item):
                continue

        new_item = deepcopy(item)
        send_to(new_item, desk_id=dest['desk'], stage_id=dest.get('stage'))

        if dest.get('macro'):
            macro = macros_service.get_macro_by_name(dest['macro'])
            try:
                macro['callback'](new_item)
            except StopDuplication:
                continue

        extra_fields = [PUBLISH_SCHEDULE, SCHEDULE_SETTINGS]
        archive_service.duplicate_content(new_item, state='routed', extra_fields=extra_fields)


def init_app(app):
    register_resource(
        NAME,
        InternalDestinationsResource,
        InternalDestinationsService,
        _app=app)

    privilege(
        name=NAME,
        label='Internal Destinations',
        description='User can manage internal destinations.')

    item_published.connect(handle_item_published)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import json

from unittest import mock
from datetime import timedelta

from superdesk.utc import utcnow
from superdesk.tests import TestCase
from superdesk.publish.formatters.ninjs_formatter import NINJSFormatter
from superdesk.publish import init_app
from bson import ObjectId


@mock.patch("superdesk.publish.subscribers.SubscribersService.generate_sequence_number", lambda self, subscriber: 1)
class NinjsFormatterTest(TestCase):
    def setUp(self):
        self.formatter = NINJSFormatter()
        init_app(self.app)
        self.maxDiff = None

    def test_text_formatter(self):
        self.app.data.insert(
            "vocabularies",
            [
                {
                    "_id": "locators",
                    "display_name": "Locators",
                    "type": "unmanageable",
                    "unique_field": "qcode",
                    "items": [
                        {
                            "is_active": True,
                            "name": "NSW",
                            "qcode": "NSW",
                            "state": "New South Wales",
                            "country": "Australia",
                            "world_region": "Oceania",
                            "group": "Australia",
                        }
                    ],
                }
            ],
        )
        embargo_ts = utcnow() + timedelta(days=2)
        article = {
            "_id": "tag:aap.com.au:20150613:12345",
            "guid": "tag:aap.com.au:20150613:12345",
            "_current_version": 1,
            "anpa_category": [{"qcode": "a"}],
            "source": "AAP",
            "headline": "This is a test headline",
            "byline": "joe",
            "slugline": "slugline",
            "subject": [
                {"qcode": "02011001", "name": "international court or tribunal", "parent": None},
                {"qcode": "02011002", "name": "extradition"},
            ],
            "anpa_take_key": "take_key",
            "unique_id": "1",
            "body_html": "The story body",
            "type": "text",
            "word_count": "1",
            "priority": 1,
            "profile": "snap",
            "state": "published",
            "urgency": 2,
            "pubstatus": "usable",
            "creditline": "sample creditline",
            "keywords": ["traffic"],
            "abstract": "<p>sample <b>abstract</b></p>",
            "place": [{"name": "NSW", "qcode": "NSW"}],
            "embargo": embargo_ts,
            "body_footer": "<p>call helpline 999 if you are planning to quit smoking</p>",
            "company_codes": [{"name": "YANCOAL AUSTRALIA LIMITED", "qcode": "YAL", "security_exchange": "ASX"}],
            "genre": [{"name": "Article", "qcode": "article"}],
            "flags": {"marked_for_legal": True},
            "extra": {"foo": "test"},
            "annotations": [{"msg": "test"}],
        }
        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        expected = {
            "guid": "tag:aap.com.au:20150613:12345",
            "version": "1",
            "place": [{"code": "NSW", "name": "New South Wales"}],
            "pubstatus": "usable",
            "body_html": "The story body<p>call helpline 999 if you are planning to quit smoking</p>",
            "type": "text",
            "subject": [
                {"code": "02011001", "name": "international court or tribunal"},
                {"code": "02011002", "name": "extradition"},
            ],
            "service": [{"code": "a"}],
            "source": "AAP",
            "headline": "This is a test headline",
            "byline": "joe",
            "urgency": 2,
            "priority": 1,
            "embargoed": embargo_ts.isoformat(),
            "profile": "snap",
            "slugline": "slugline",
            "description_text": "sample abstract",
            "description_html": "<p>sample <b>abstract</b></p>",
            "keywords": ["traffic"],
            "organisation": [
                {
                    "name": "YANCOAL AUSTRALIA LIMITED",
                    "rel": "Securities Identifier",
                    "symbols": [{"ticker": "YAL", "exchange": "ASX"}],
                }
            ],
            "genre": [{"name": "Article", "code": "article"}],
            "signal": [{"name": "Content Warning", "code": "cwarn", "scheme": "http://cv.iptc.org/newscodes/signal/"}],
            "extra": {"foo": "test"},
            "charcount": 67,
            "wordcount": 13,
            "readtime": 0,
            "annotations": article["annotations"],
        }
        self.assertEqual(json.loads(doc), expected)

    def test_picture_formatter(self):
        article = {
            "guid": "20150723001158606583",
            "_current_version": 1,
            "slugline": "AMAZING PICTURE",
            "original_source": "AAP",
            "renditions": {
                "viewImage": {
                    "width": 640,
                    "href": "http://localhost:5000/api/upload/55b032041d41c8d278d21b6f/raw?_schema=http",
                    "mimetype": "image/jpeg",
                    "height": 401,
                },
                "original": {
                    "href": "https://one-api.aap.com.au/api/v3/Assets/20150723001158606583/Original/download",
                    "mimetype": "image/jpeg",
                },
            },
            "byline": "MICKEY MOUSE",
            "headline": "AMAZING PICTURE",
            "versioncreated": "2015-07-23T00:15:00.000Z",
            "ednote": "TEST ONLY",
            "type": "picture",
            "pubstatus": "usable",
            "source": "AAP",
            "description": "The most amazing picture you will ever see",
            "guid": "20150723001158606583",
            "body_footer": "<p>call helpline 999 if you are planning to quit smoking</p>",
        }
        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        expected = {
            "byline": "MICKEY MOUSE",
            "renditions": {
                "original": {
                    "href": "https://one-api.aap.com.au/api/v3/Assets/20150723001158606583/Original/download",
                    "mimetype": "image/jpeg",
                }
            },
            "headline": "AMAZING PICTURE",
            "pubstatus": "usable",
            "version": "1",
            "versioncreated": "2015-07-23T00:15:00.000Z",
            "guid": "20150723001158606583",
            "description_html": "The most amazing picture you will ever see<p>call helpline 999 if you are planning to "
            "quit smoking</p>",
            "type": "picture",
            "priority": 5,
            "slugline": "AMAZING PICTURE",
            "ednote": "TEST ONLY",
            "source": "AAP",
        }
        self.assertEqual(expected, json.loads(doc))

    def test_composite_formatter(self):
        article = {
            "guid": "urn:newsml:localhost:2015-07-24T15:05:00.116047:435c93c2-492c-4668-ab47-ae6e2b9b1c2c",
            "groups": [
                {"id": "root", "refs": [{"idRef": "main"}, {"idRef": "sidebars"}], "role": "grpRole:NEP"},
                {
                    "id": "main",
                    "refs": [
                        {
                            "renditions": {},
                            "slugline": "Boat",
                            "guid": "tag:localhost:2015:515b895a-b336-48b2-a506-5ffaf561b916",
                            "headline": "WA:Navy steps in with WA asylum-seeker boat",
                            "location": "archive",
                            "type": "text",
                            "itemClass": "icls:text",
                            "residRef": "tag:localhost:2015:515b895a-b336-48b2-a506-5ffaf561b916",
                        }
                    ],
                    "role": "grpRole:main",
                },
                {
                    "id": "sidebars",
                    "refs": [
                        {
                            "renditions": {
                                "original_source": {
                                    "href": "https://one-api.aap.com.au\
                                        /api/v3/Assets/20150723001158639795/Original/download",
                                    "mimetype": "image/jpeg",
                                },
                                "original": {
                                    "width": 2784,
                                    "height": 4176,
                                    "href": "http://localhost:5000\
                                    /api/upload/55b078b21d41c8e974d17ec5/raw?_schema=http",
                                    "mimetype": "image/jpeg",
                                    "media": "55b078b21d41c8e974d17ec5",
                                },
                                "thumbnail": {
                                    "width": 80,
                                    "height": 120,
                                    "href": "http://localhost:5000\
                                    /api/upload/55b078b41d41c8e974d17ed3/raw?_schema=http",
                                    "mimetype": "image/jpeg",
                                    "media": "55b078b41d41c8e974d17ed3",
                                },
                                "viewImage": {
                                    "width": 426,
                                    "height": 640,
                                    "href": "http://localhost:5000\
                                    /api/upload/55b078b31d41c8e974d17ed1/raw?_schema=http",
                                    "mimetype": "image/jpeg",
                                    "media": "55b078b31d41c8e974d17ed1",
                                },
                                "baseImage": {
                                    "width": 933,
                                    "height": 1400,
                                    "href": "http://localhost:5000\
                                    /api/upload/55b078b31d41c8e974d17ecf/raw?_schema=http",
                                    "mimetype": "image/jpeg",
                                    "media": "55b078b31d41c8e974d17ecf",
                                },
                            },
                            "slugline": "ABC SHOP CLOSURES",
                            "type": "picture",
                            "guid": "urn:newsml:localhost:2015-07-24T15:04:29.589984:"
                            "af3bef9a-5002-492b-a15a-8b460e69b164",
                            "headline": "ABC SHOP CLOSURES",
                            "location": "archive",
                            "itemClass": "icls:picture",
                            "residRef": "urn:newsml:localhost:2015-07-24T15:04:29.589984:"
                            "af3bef9a-5002-492b-a15a-8b460e69b164",
                        }
                    ],
                    "role": "grpRole:sidebars",
                },
            ],
            "description": "",
            "operation": "update",
            "sign_off": "mar",
            "type": "composite",
            "pubstatus": "usable",
            "version_creator": "558379451d41c83ff598a3af",
            "language": "en",
            "guid": "urn:newsml:localhost:2015-07-24T15:05:00.116047:435c93c2-492c-4668-ab47-ae6e2b9b1c2c",
            "unique_name": "#145",
            "headline": "WA:Navy steps in with WA asylum-seeker boat",
            "original_creator": "558379451d41c83ff598a3af",
            "source": "AAP",
            "_etag": "b41df79084304219524a092abf07ecba9e1bb2c5",
            "slugline": "Boat",
            "firstcreated": "2015-07-24T05:05:00.000Z",
            "unique_id": 145,
            "versioncreated": "2015-07-24T05:05:14.000Z",
            "_updated": "2015-07-24T05:05:25.000Z",
            "family_id": "urn:newsml:localhost:2015-07-24T15:05:00.116047:435c93c2-492c-4668-ab47-ae6e2b9b1c2c",
            "_current_version": 2,
            "_created": "2015-07-24T05:05:00.000Z",
            "version": 2,
        }

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        expected = {
            "headline": "WA:Navy steps in with WA asylum-seeker boat",
            "version": "2",
            "guid": "urn:newsml:localhost:2015-07-24T15:05:00.116047:435c93c2-492c-4668-ab47-ae6e2b9b1c2c",
            "associations": {
                "main": {"guid": "tag:localhost:2015:515b895a-b336-48b2-a506-5ffaf561b916", "type": "text"},
                "sidebars": {
                    "guid": "urn:newsml:localhost:2015-07-24T15:04:29.589984:af3bef9a-5002-492b-a15a-8b460e69b164",
                    "type": "picture",
                },
            },
            "firstcreated": "2015-07-24T05:05:00.000Z",
            "versioncreated": "2015-07-24T05:05:14.000Z",
            "type": "composite",
            "pubstatus": "usable",
            "language": "en",
            "priority": 5,
            "slugline": "Boat",
            "source": "AAP",
        }
        self.assertEqual(expected, json.loads(doc))

    def test_item_with_usable_associations(self):
        article = {
            "_id": "urn:bar",
            "guid": "urn:bar",
            "_current_version": 1,
            "type": "text",
            "associations": {
                "image": {
                    "_id": "urn:foo",
                    "guid": "urn:foo",
                    "pubstatus": "usable",
                    "headline": "Foo",
                    "type": "picture",
                    "task": {},
                    "copyrightholder": "Foo ltd.",
                    "description_text": "Foo picture",
                    "renditions": {
                        "original": {
                            "href": "http://example.com",
                            "width": 100,
                            "height": 80,
                            "mimetype": "image/jpeg",
                            "CropLeft": 0,
                        }
                    },
                }
            },
        }

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        formatted = json.loads(doc)
        self.assertIn("associations", formatted)
        self.assertIn("image", formatted["associations"])
        image = formatted["associations"]["image"]
        self.assertEqual("urn:foo", image["guid"])
        self.assertEqual("Foo", image["headline"])
        self.assertEqual("usable", image["pubstatus"])
        self.assertNotIn("task", image)
        self.assertEqual("Foo ltd.", image["copyrightholder"])
        self.assertEqual("Foo picture", image["description_text"])
        rendition = image["renditions"]["original"]
        self.assertEqual(100, rendition["width"])
        self.assertEqual(80, rendition["height"])
        self.assertEqual("image/jpeg", rendition["mimetype"])
        self.assertNotIn("CropLeft", rendition)

    def test_item_with_empty_associations(self):
        article = {
            "_id": "urn:bar",
            "guid": "urn:bar",
            "_current_version": 1,
            "type": "text",
            "associations": {"image": None},
        }

        _, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        formatted = json.loads(doc)
        self.assertIn("associations", formatted)
        self.assertNotIn("image", formatted["associations"])

    def test_vidible_formatting(self):
        article = {
            "_id": "tag:aap.com.au:20150613:12345",
            "guid": "tag:aap.com.au:20150613:12345",
            "_current_version": 1,
            "source": "AAP",
            "headline": "This is a test headline",
            "slugline": "slugline",
            "unique_id": "1",
            "body_html": "The story body",
            "type": "text",
            "state": "published",
            "pubstatus": "usable",
            "associations": {
                "embedded5346670761": {
                    "uri": "56ba77bde4b0568f54a1ce68",
                    "alt_text": "alternative",
                    "copyrightholder": "Edouard",
                    "copyrightnotice": "Edited with Gimp",
                    "usageterms": "indefinite-usage",
                    "type": "video",
                    "title": "Embed title",
                    "company": "Press Association",
                    "url": "https://videos.vidible.tv/prod/2016-02/09/56ba777ce4b0b6448ed478f5_640x360.mp4",
                    "thumbnail": "https://cdn-ssl.vidible.tv/2016-02/09/56ba777ce4b0b6448ed478f5_60x60.jpg",
                    "duration": 100,
                    "width": 400,
                    "height": 200,
                }
            },
        }
        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        expected = {
            "guid": "tag:aap.com.au:20150613:12345",
            "version": "1",
            "pubstatus": "usable",
            "body_html": "The story body",
            "type": "text",
            "headline": "This is a test headline",
            "slugline": "slugline",
            "priority": 5,
            "source": "AAP",
            "charcount": 14,
            "wordcount": 3,
            "readtime": 0,
            "associations": {
                "embedded5346670761": {
                    "guid": "56ba77bde4b0568f54a1ce68",
                    "type": "video",
                    "version": "1",
                    "priority": 5,
                    "body_text": "alternative",
                    "copyrightholder": "Edouard",
                    "copyrightnotice": "Edited with Gimp",
                    "usageterms": "indefinite-usage",
                    "headline": "Embed title",
                    "organisation": [{"name": "Press Association"}],
                    "renditions": {
                        "original": {
                            "href": "https://videos.vidible.tv/prod/2016-02/09/56ba777ce4b0b6448ed478f5_640x360.mp4",
                            "duration": 100,
                            "width": 400,
                            "height": 200,
                        },
                        "thumbnail": {
                            "href": "https://cdn-ssl.vidible.tv/2016-02/09/56ba777ce4b0b6448ed478f5_60x60.jpg"
                        },
                    },
                }
            },
        }
        self.assertEqual(json.loads(doc), expected)

    def test_copyright_holder_notice(self):
        self.app.data.insert(
            "vocabularies",
            [
                {
                    "_id": "rightsinfo",
                    "items": [
                        {
                            "is_active": True,
                            "name": "default",
                            "copyrightHolder": "copyright holder",
                            "copyrightNotice": "copyright notice",
                            "usageTerms": "",
                        }
                    ],
                }
            ],
        )

        article = {"_id": "urn:bar", "_current_version": 1, "guid": "urn:bar", "type": "text"}

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        data = json.loads(doc)

        self.assertEqual("copyright holder", data["copyrightholder"])
        self.assertEqual("copyright notice", data["copyrightnotice"])
        self.assertEqual("", data["usageterms"])

    def test_body_html(self):
        article = {
            "_id": "urn:bar",
            "_current_version": 1,
            "guid": "urn:bar",
            "type": "text",
            "body_html": (250 * 6 - 40) * "word ",
        }

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        data = json.loads(doc)

        self.assertEqual(data["charcount"], 7300)
        self.assertEqual(data["wordcount"], 1460)
        self.assertEqual(data["readtime"], 6)

    def test_body_text(self):
        article = {
            "_id": "urn:bar",
            "_current_version": 1,
            "guid": "urn:bar",
            "type": "text",
            "body_text": (250 * 7 - 40) * "word ",
        }

        data = self._format(article)

        self.assertEqual(data["charcount"], 8550)
        self.assertEqual(data["wordcount"], 1710)
        self.assertEqual(data["readtime"], 7)

        # check japanese
        article["language"] = "ja"
        article["body_text"] = 5000 * "x"
        data = self._format(article)
        self.assertEqual(data["readtime"], 8)

        article["body_text"] = 5000 * " "
        data = self._format(article)
        self.assertEqual(data["readtime"], 0)

    def _format(self, article):
        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        return json.loads(doc)

    def test_empty_amstract(self):
        article = {"_id": "urn:bar", "_current_version": 1, "guid": "urn:bar", "type": "text", "abstract": ""}

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        data = json.loads(doc)

        self.assertEqual(data["description_html"], "")
        self.assertEqual(data["description_text"], "")

    def test_authors(self):
        self.app.data.insert(
            "users",
            [
                {
                    "_id": "test_id",
                    "username": "author 1",
                    "display_name": "author 1",
                    "is_author": True,
                    "job_title": "writer_code",
                    "biography": "bio 1",
                    "facebook": "johnsmith",
                    "twitter": "@smith_john",
                    "instagram": "john_s",
                    "picture_url": "http://example.com",
                },
                {
                    "_id": "test_id_2",
                    "username": "author 2",
                    "display_name": "author 2",
                    "is_author": True,
                    "job_title": "reporter_code",
                    "biography": "bio 2",
                },
            ],
        )

        self.app.data.insert(
            "vocabularies",
            [
                {
                    "_id": "job_titles",
                    "display_name": "Job Titles",
                    "type": "manageable",
                    "unique_field": "qcode",
                    "items": [
                        {"is_active": True, "name": "Writer", "qcode": "writer_code"},
                        {"is_active": True, "name": "Reporter", "qcode": "reporter_code"},
                    ],
                    "schema": {"name": {}, "qcode": {}},
                }
            ],
        )

        article = {
            "_id": "urn:bar",
            "_current_version": 1,
            "guid": "urn:bar",
            "type": "text",
            "authors": [
                {
                    "_id": ["test_id", "writer"],
                    "role": "writer",
                    "name": "Writer",
                    "parent": "test_id",
                    "sub_label": "author 1",
                },
                {
                    "_id": ["test_id_2", "writer"],
                    "role": "photographer",
                    "name": "photographer",
                    "parent": "test_id_2",
                    "sub_label": "author 2",
                },
            ],
        }

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        data = json.loads(doc)

        expected = [
            {
                "name": "author 1",
                "role": "writer",
                "jobtitle": {"qcode": "writer_code", "name": "Writer"},
                "biography": "bio 1",
                "facebook": "johnsmith",
                "twitter": "@smith_john",
                "instagram": "john_s",
                "avatar_url": "http://example.com",
            },
            {
                "name": "author 2",
                "role": "photographer",
                "jobtitle": {"qcode": "reporter_code", "name": "Reporter"},
                "biography": "bio 2",
            },
        ]
        self.assertEqual(data["authors"], expected)

    def test_author_missing_parent(self):
        """Test that older items with missing parent don't make the formatter crashing"""
        article = {
            "_id": "urn:bar",
            "_current_version": 1,
            "guid": "urn:bar",
            "type": "text",
            "authors": [
                {"_id": ["test_id", "writer"], "role": "writer", "name": "Writer", "sub_label": "author 1"},
                {
                    "_id": ["test_id_2", "writer"],
                    "role": "photographer",
                    "name": "photographer",
                    "sub_label": "author 2",
                },
            ],
        }

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        data = json.loads(doc)

        expected = {
            "guid": "urn:bar",
            "version": "1",
            "type": "text",
            "priority": 5,
            "authors": [
                {"name": "Writer", "role": "writer", "biography": ""},
                {"name": "photographer", "role": "photographer", "biography": ""},
            ],
        }

        self.assertEqual(data, expected)

    def test_place(self):
        self.app.data.insert(
            "vocabularies",
            [
                {
                    "_id": "locators",
                    "display_name": "Locators",
                    "type": "unmanageable",
                    "unique_field": "qcode",
                    "items": [
                        {
                            "is_active": True,
                            "name": "JPN",
                            "qcode": "JPN",
                            "state": "",
                            "country": "Japan",
                            "world_region": "Asia",
                            "group": "Rest Of World",
                        },
                        {"is_active": True, "name": "SAM", "qcode": "SAM", "group": "Rest Of World"},
                        {
                            "is_active": True,
                            "name": "UK",
                            "qcode": "UK",
                            "state": "",
                            "country": "",
                            "world_region": "Europe",
                            "group": "Rest Of World",
                        },
                    ],
                }
            ],
        )
        article = {
            "_id": "urn:bar",
            "_current_version": 1,
            "guid": "urn:bar",
            "type": "text",
            "place": [{"name": "JPN", "qcode": "JPN"}],
        }

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        data = json.loads(doc)

        self.assertEqual(data["place"], [{"code": "JPN", "name": "Japan"}])

        article = {
            "_id": "urn:bar",
            "_current_version": 1,
            "guid": "urn:bar",
            "type": "text",
            "place": [{"name": "SAM", "qcode": "SAM"}],
        }

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        data = json.loads(doc)

        self.assertEqual(data["place"], [{"code": "SAM", "name": "Rest Of World"}])

        article = {
            "_id": "urn:bar",
            "_current_version": 1,
            "guid": "urn:bar",
            "type": "text",
            "place": [{"name": "UK", "qcode": "UK"}],
        }

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        data = json.loads(doc)

        self.assertEqual(data["place"], [{"code": "UK", "name": "Europe"}])

    def test_translations(self):
        """Check that fields are correctly translated"""
        article = {
            "_id": "5a68a134cc3a2d4bd6399177",
            "type": "text",
            "guid": "test",
            "genre": [
                {
                    "name": "Education",
                    "qcode": "genre_custom:Education",
                    "translations": {
                        "name": {"de": "Weiterbildung", "it": "Educazione finanziaria", "ja": "トレーニング用教材"}
                    },
                    "scheme": "genre_custom",
                }
            ],
            "language": "ja",
            "headline": "test",
            "body_html": "<p>test ter</p>",
            "subject": [
                {
                    "name": "Outcome orientated solutions",
                    "parent": "subject:01000000",
                    "qcode": "subject:01000002",
                    "translations": {
                        "name": {"de": "Ergebnisorientiert", "it": "Orientato ai risultati ", "ja": "アウトカム・オリエンティッド"}
                    },
                    "scheme": "subject_custom",
                },
                {
                    "name": "Austria",
                    "qcode": "country_custom:1001002",
                    "translations": {"name": {"de": "\u00d6sterreich", "it": "Austria", "ja": "オーストリア"}},
                    "scheme": "country_custom",
                },
                {
                    "name": "Asia ex Japan",
                    "qcode": "region_custom:Asia ex Japan",
                    "translations": {"name": {"de": "Asien exkl. Japan", "it": "Asia escl. Giappone", "ja": "日本除くアジア"}},
                    "scheme": "region_custom",
                },
                {"name": "no translations", "qcode": "test", "translations": None, "scheme": "test"},
            ],
        }
        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        ninjs = json.loads(doc)
        expected_genre = [{"code": "genre_custom:Education", "name": "トレーニング用教材", "scheme": "genre_custom"}]
        self.assertEqual(ninjs["genre"], expected_genre)
        expected_subject = [
            {"code": "subject:01000002", "name": "アウトカム・オリエンティッド", "scheme": "subject_custom"},
            {"code": "country_custom:1001002", "name": "オーストリア", "scheme": "country_custom"},
            {"code": "region_custom:Asia ex Japan", "name": "日本除くアジア", "scheme": "region_custom"},
            {"code": "test", "name": "no translations", "scheme": "test"},
        ]
        self.assertEqual(ninjs["subject"], expected_subject)

    def test_place_geonames(self):
        article = {
            "_id": "urn:bar",
            "_current_version": 1,
            "guid": "urn:bar",
            "type": "text",
            "place": [
                {
                    "name": "Kobeřice",
                    "code": "3073493",
                    "scheme": "geonames",
                    "state": "Moravskoslezský kraj",
                    "country": "Česko",
                    "state_code": "80",
                    "country_code": "CZ",
                    "location": {"lat": 49.98548, "lon": 18.05212},
                }
            ],
        }

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        ninjs = json.loads(doc)

        self.assertEqual({"name": "Kobeřice", "code": "3073493", "scheme": "geonames"}, ninjs["place"][0])

    def test_custom_media(self):
        """Test that custom media are put in "groups" field and not associations (SDESK-2955)"""
        self.app.data.insert(
            "content_types",
            [
                {
                    "_id": ObjectId("5ba11fec0d6f1301ac3cbd13"),
                    "label": "custom media field multi",
                    "editor": {
                        "slugline": {"order": 2, "sdWidth": "full"},
                        "headline": {"order": 3, "formatOptions": ["underline", "link", "bold"]},
                        "custom_media_field_multi_1": {"order": 1},
                    },
                    "schema": {
                        "headline": {"type": "string", "required": False, "maxlength": 64, "nullable": True},
                        "slugline": {"type": "string", "required": False, "maxlength": 24, "nullable": True},
                        "custom_media_field_multi_1": {
                            "type": "media",
                            "required": False,
                            "enabled": True,
                            "nullable": True,
                        },
                    },
                }
            ],
        )
        article = {
            "_id": "5ba1224e0d6f13056bd82d50",
            "type": "text",
            "version": 1,
            "profile": "5ba11fec0d6f1301ac3cbd13",
            "format": "HTML",
            "template": "5ba11fec0d6f1301ac3cbd15",
            "headline": "custom media field multi",
            "slugline": "test custom media2",
            "guid": "123",
            "associations": {
                "custom_media_field_multi_1--1": {
                    "renditions": {
                        "original": {
                            "href": "http://localhost:5000/api/upload-raw/123.jpg",
                            "media": "abc",
                            "mimetype": "image/jpeg",
                            "width": 550,
                            "height": 331,
                        }
                    },
                    "media": "abc",
                    "type": "picture",
                    "guid": "tag:localhost:5000:2018:3710ef88-9567-4dbb-a96b-cb53df15b66e",
                },
                "custom_media_field_multi_1--2": {
                    "renditions": {
                        "original": {
                            "href": "http://localhost:5000/api/upload-raw/456.jpg",
                            "media": "cde",
                            "mimetype": "image/jpeg",
                            "width": 550,
                            "height": 331,
                        }
                    },
                    "media": "cde",
                    "type": "picture",
                    "guid": "tag:localhost:5000:2018:3710ef88-9567-4dbb-a96b-cb53df15b66e",
                },
            },
        }

        expected = {
            "associations": {
                "custom_media_field_multi_1--1": {
                    "guid": "tag:localhost:5000:2018:3710ef88-9567-4dbb-a96b-cb53df15b66e",
                    "priority": 5,
                    "renditions": {
                        "original": {
                            "height": 331,
                            "href": "http://localhost:5000/api/upload-raw/123.jpg",
                            "media": "abc",
                            "mimetype": "image/jpeg",
                            "width": 550,
                        }
                    },
                    "type": "picture",
                    "version": "1",
                },
                "custom_media_field_multi_1--2": {
                    "guid": "tag:localhost:5000:2018:3710ef88-9567-4dbb-a96b-cb53df15b66e",
                    "priority": 5,
                    "renditions": {
                        "original": {
                            "height": 331,
                            "href": "http://localhost:5000/api/upload-raw/456.jpg",
                            "media": "cde",
                            "mimetype": "image/jpeg",
                            "width": 550,
                        }
                    },
                    "type": "picture",
                    "version": "1",
                },
            },
            "extra_items": {
                "custom_media_field_multi_1": {
                    "items": [
                        {
                            "guid": "tag:localhost:5000:2018:3710ef88-9567-4dbb-a96b-cb53df15b66e",
                            "priority": 5,
                            "renditions": {
                                "original": {
                                    "height": 331,
                                    "href": "http://localhost:5000/api/upload-raw/123.jpg",
                                    "media": "abc",
                                    "mimetype": "image/jpeg",
                                    "width": 550,
                                }
                            },
                            "type": "picture",
                            "version": "1",
                        },
                        {
                            "guid": "tag:localhost:5000:2018:3710ef88-9567-4dbb-a96b-cb53df15b66e",
                            "priority": 5,
                            "renditions": {
                                "original": {
                                    "height": 331,
                                    "href": "http://localhost:5000/api/upload-raw/456.jpg",
                                    "media": "cde",
                                    "mimetype": "image/jpeg",
                                    "width": 550,
                                }
                            },
                            "type": "picture",
                            "version": "1",
                        },
                    ]
                }
            },
            "guid": "123",
            "headline": "custom media field multi",
            "priority": 5,
            "profile": "custommediafieldmulti",
            "slugline": "test custom media2",
            "type": "text",
            "version": "1",
        }

        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        ninjs = json.loads(doc)
        self.assertEqual(ninjs, expected)



from lxml import etree

from ..base import BasePackageElement


class Story(BasePackageElement):
    """
    Story which represents `p`, `h1`, `h2`, `h3`, `h4`, `h5`, `h6`, `blockquote`, `pre`, `headline`, `byline` html tags.

    As Adobe's idml specification explains:
    ``
    A story, or “text flow” is the basic text container in an InDesign document;
    all text exists inside stories. Stories are associated with at least one text frame or text path,
    and can span any number of linked text frames or text paths in a document.
    ``
    """

    STORY_DEFAULTS = {
        'AppliedNamedGrid': 'n',
        'AppliedTOCStyle': 'n',
        'TrackChanges': 'false',
        'StoryTitle': '$ID/'
    }
    STORYPREFERENCE_DEFAULTS = {
        'OpticalMarginAlignment': 'false',
        'OpticalMarginSize': '12',
        'FrameType': 'TextFrameType',
        'StoryOrientation': 'Horizontal',
        'StoryDirection': 'LeftToRightDirection'
    }
    HYPERLINKTEXTSOURCE_DEFAULTS = {
        'Hidden': 'false',
        'AppliedCharacterStyle': 'CharacterStyle/$ID/Hyperlink'
    }
    PARAGRAPHSTYLERANGE_DEFAULTS = {
        'AppliedParagraphStyle': 'ParagraphStyle/$ID/NormalParagraphStyle'
    }
    CHARACTERSTYLERANGE_DEFAULTS = {
        'AppliedCharacterStyle': 'CharacterStyle/$ID/[No character style]',
    }
    BLOCK_TAGS_MAPPING = {
        'p': {
            'markup_tag': 'NormalParagraph',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Paragraphs%3aNormalParagraph'
            },
        },
        'blockquote': {
            'markup_tag': 'Blockquote',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Paragraphs%3aBlockquote'
            },
        },
        'pre': {
            'markup_tag': 'Preformatted',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Paragraphs%3aPreformatted'
            },
        },
        'h1': {
            'markup_tag': 'Heading1',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Headings%3aHeading1'
            },
        },
        'h2': {
            'markup_tag': 'Heading2',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Headings%3aHeading2'
            },
        },
        'h3': {
            'markup_tag': 'Heading3',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Headings%3aHeading3'
            },
        },
        'h4': {
            'markup_tag': 'Heading4',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Headings%3aHeading4'
            },
        },
        'h5': {
            'markup_tag': 'Heading5',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Headings%3aHeading5'
            },
        },
        'h6': {
            'markup_tag': 'Heading6',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Headings%3aHeading6'
            },
        },
        'table': {
            'markup_tag': 'Table',
            'Table': {
                'AppliedTableStyle': 'TableStyle/$ID/NormalTable'
            },
        },
        'ul': {
            'markup_tag': 'UnorderedList',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Lists%3aUnorderedList',
                'BulletsAndNumberingListType': 'BulletList',
            },
        },
        'ol': {
            'markup_tag': 'OrderedList',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Lists%3aOrderedList',
                'BulletsAndNumberingListType': 'NumberedList',
            },
        },
        # custom
        'headline': {
            'markup_tag': 'Headline',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Custom%3aHeadline'
            },
        },
        'byline': {
            'markup_tag': 'Byline',
            'ParagraphStyleRange': {
                'AppliedParagraphStyle': 'ParagraphStyle/Custom%3aByline'
            },
        },
    }
    INLINE_TAGS_MAPPING = {
        'b': [
            {
                'attribute': 'FontStyle',
                'value': 'Bold',
                'appendable': True
            }
        ],
        'i': [
            {
                'attribute': 'FontStyle',
                'value': 'Italic',
                'appendable': True
            }
        ],
        'u': [
            {
                'attribute': 'Underline',
                'value': 'true',
                'appendable': False
            }
        ],
        's': [
            {
                'attribute': 'StrikeThru',
                'value': 'true',
                'appendable': False
            }
        ],
        'sub': [
            {
                'attribute': 'Position',
                'value': 'Subscript',
                'appendable': False
            }
        ],
        'sup': [
            {
                'attribute': 'Position',
                'value': 'Superscript',
                'appendable': False
            }
        ],
        'a': [
            {
                'attribute': 'AppliedCharacterStyle',
                'value': 'CharacterStyle/$ID/Hyperlink',
                'appendable': False
            }
        ]
    }

    def __init__(self, self_id, element, attributes=None, markup_tag=None):
        self.self_id = self_id
        self._element = element
        self._markup_tag = markup_tag
        self.links = []
        super().__init__(attributes)

    @property
    def filename(self):
        """
        Filename inside IDML package.
        Used as a filename for a file inside zip container.
        :return str: filename
        """
        return 'Stories/Story_{}.xml'.format(self.self_id)

    def _build_etree(self):
        self._etree = etree.Element(
            etree.QName(self.XMLNS_IDPKG, 'Story'),
            nsmap={'idPkg': self.XMLNS_IDPKG}
        )
        self._etree.set('DOMVersion', self.DOM_VERSION)
        self._add_story()

    def _add_story(self):
        """
        Create <Story..> tag
        :return lxml.etree: Story etree instance
        """
        # merge Story attributes
        story_attributes = self.merge_attributes(
            self.STORY_DEFAULTS,
            self._attributes.get('Story', {})
        )
        story_attributes.update({'Self': self.self_id})
        # Story
        story = etree.SubElement(
            self._etree,
            'Story',
            attrib=story_attributes
        )
        # StoryPreference
        etree.SubElement(
            story,
            'StoryPreference',
            attrib=self.merge_attributes(
                self.STORYPREFERENCE_DEFAULTS,
                self._attributes.get('StoryPreference', {})
            )
        )

        if self._markup_tag:
            # XMLElement to tag a story
            paragraphstylerange_container = etree.SubElement(
                story,
                'XMLElement',
                attrib={
                    'Self': '{}_{}'.format(self.self_id, self._markup_tag.lower()),
                    'XMLContent': self.self_id,
                    'MarkupTag': 'XMLTag/{}'.format(self._markup_tag)
                }
            )
        else:
            paragraphstylerange_container = story

        # ParagraphStyleRange
        paragraphstylerange = etree.SubElement(
            paragraphstylerange_container,
            'ParagraphStyleRange',
            attrib=self.merge_attributes(
                self.PARAGRAPHSTYLERANGE_DEFAULTS,
                self._attributes.get('ParagraphStyleRange', {})
            )
        )

        paragraphstylerange[:] = self._handle_inline_tags(self._element)

        return story

    def _handle_inline_tags(self, element):
        """
        Convert html element's inline style to idml representation.
        :param lxml.etree element: html element
        :return list: list of CharacterStyleRange lxml.etree instances
        """
        element_dict = self._etree_element_to_dict(element)
        return self._create_characterstylerange_recursive(element_dict)

    def _etree_element_to_dict(self, element, parents=None):
        """
        Converts html element to dict.
        Example:
            `<p>normal <b>bold <u>bold-unerline</u> bold</b> normal</p>`
            will be converted to:
            ``
            {
                'attrib': {},
                'childs': (
                    {
                        'attrib': {},
                        'childs': (
                            {
                                'attrib': {},
                                'childs': (),
                                'parents': ('p', 'b'),
                                'tag': 'u',
                                'tail': ' bold',
                                'text': 'bold-unerline',
                            },
                        ),
                        'parents': ('p', ),
                        'tag': 'b',
                        'tail': ' normal',
                        'text': 'bold ',
                    },
                ),
                'parents': (),
                'tag': 'p',
                'tail': None,
                'text': 'normal ',
            }
            ``
        :param lxml.etree element: html element
        :param list parents: list of parent tags
        :return dict: dictionary
        """
        if not parents:
            parents = []

        element_dict = {
            'tag': element.tag,
            'attrib': element.attrib,
            'parents': tuple(parents),
            'text': element.text,
            'tail': element.tail
        }
        parents.append(element.tag)
        element_dict['childs'] = tuple(
            self._etree_element_to_dict(
                child,
                list(parents)
            ) for child in element.iterchildren()
        )

        return element_dict

    def _create_characterstylerange_recursive(self, element_dict, tail=False):
        """
        Create CharacterStyleRange(s) recursively.
        Example:
        ``
        <ParagraphStyleRange ...>
        <CharacterStyleRange AppliedCharacterStyle="..." FontStyle="Bold">
          <Content>Curabitur </Content>
        </CharacterStyleRange>
        <CharacterStyleRange AppliedCharacterStyle="..." FontStyle="Bold Italic">
          <Content>arcu erat</Content>
        </CharacterStyleRange>
        <CharacterStyleRange AppliedCharacterStyle="..." FontStyle="Bold">
          <Content>, accumsan </Content>
        </CharacterStyleRange>
        <CharacterStyleRange AppliedCharacterStyle="..." FontStyle="Bold" Underline="true">
          <Content>id imperdiet et</Content>
        </CharacterStyleRange>
        ...
        </ParagraphStyleRange>
        ``
        :param dict element_dict: dict representation of html element
        :param bool tail: if element has tail
        :return:
        """

        characterstylerange_list = []

        def apply_tag_style(characterstylerange, tag):
            for rule in self.INLINE_TAGS_MAPPING[tag]:
                attribute = rule['attribute']
                value = rule['value']
                appendable = rule['appendable']

                if characterstylerange.get(attribute) and appendable:
                    if attribute == 'FontStyle' and value == 'Bold':
                        characterstylerange.set(
                            attribute,
                            '{} {}'.format(value, characterstylerange.get(attribute))
                        )
                    else:
                        characterstylerange.set(
                            attribute,
                            '{} {}'.format(characterstylerange.get(attribute), value)
                        )
                else:
                    characterstylerange.set(attribute, value)

        # text
        if element_dict['text'] and len(element_dict['text']):
            characterstylerange_list.append(
                etree.Element('CharacterStyleRange', attrib=self.CHARACTERSTYLERANGE_DEFAULTS)
            )
            # apply parents styles
            for tag in element_dict['parents']:
                if self.INLINE_TAGS_MAPPING.get(tag):
                    apply_tag_style(characterstylerange_list[-1], tag)

            # apply own style
            tag = element_dict['tag']
            if self.INLINE_TAGS_MAPPING.get(tag):
                apply_tag_style(characterstylerange_list[-1], tag)

            content = etree.Element('Content')
            content.text = element_dict['text']

            if tag == 'a':
                self.links.append({
                    'self_id': '{}_{}'.format(self.self_id, len(self.links) + 1),
                    'text': element_dict['text'],
                    'href': element_dict['attrib'].get('href')
                })

                hyperlinktextsource = etree.SubElement(
                    characterstylerange_list[-1],
                    'HyperlinkTextSource',
                    attrib=self.merge_attributes(
                        self.HYPERLINKTEXTSOURCE_DEFAULTS,
                        {
                            'Self': self.links[-1]['self_id'],
                            'Name': 'Hyperlink {}'.format(self.links[-1]['self_id']),
                        }
                    )
                )
                hyperlinktextsource.append(content)
            else:
                characterstylerange_list[-1].append(content)

        # childs
        for child in element_dict['childs']:
            characterstylerange_list += self._create_characterstylerange_recursive(child, tail=True)

        # tail
        if tail and element_dict['tail'] and len(element_dict['tail']):
            characterstylerange_list.append(
                etree.Element(
                    'CharacterStyleRange',
                    attrib=self.CHARACTERSTYLERANGE_DEFAULTS
                )
            )
            # apply parents styles
            for tag in element_dict['parents']:
                if self.INLINE_TAGS_MAPPING.get(tag):
                    apply_tag_style(characterstylerange_list[-1], tag)

            content = etree.SubElement(
                characterstylerange_list[-1],
                'Content'
            )
            content.text = element_dict['tail']

        return characterstylerange_list

    @property
    def length(self):
        return len(" ".join(etree.XPath(".//text()")(self._element)))

    @staticmethod
    def guess_height(story, inner_width):
        style = story._etree.xpath('.//ParagraphStyleRange')[-1].get('AppliedParagraphStyle')
        if not style:
            return 100

        style = style.rsplit('%3a', 1)[-1]
        style_map = {
            'Headline': {
                'PointSize': 48,
                'Leading': 48,
                'LeftIndent': 0,
                'RightIndent': 0
            },
            'Byline': {
                'PointSize': 20,
                'Leading': 20,
                'LeftIndent': 0,
                'RightIndent': 0
            },
            'Heading1': {
                'PointSize': 40,
                'Leading': 40,
                'LeftIndent': 0,
                'RightIndent': 0
            },
            'Heading2': {
                'PointSize': 30,
                'Leading': 30,
                'LeftIndent': 0,
                'RightIndent': 0
            },
            'Heading3': {
                'PointSize': 20,
                'Leading': 20,
                'LeftIndent': 0,
                'RightIndent': 0
            },
            'Heading4': {
                'PointSize': 14,
                'Leading': 14,
                'LeftIndent': 0,
                'RightIndent': 0
            },
            'Heading5': {
                'PointSize': 11,
                'Leading': 11,
                'LeftIndent': 0,
                'RightIndent': 0
            },
            'Heading6': {
                'PointSize': 9,
                'Leading': 9,
                'LeftIndent': 0,
                'RightIndent': 0
            },
            'Blockquote': {
                'PointSize': 12,
                'Leading': 12,
                'LeftIndent': 30,
                'RightIndent': 30
            },
            'default': {
                'PointSize': 12,
                'Leading': 12,
                'LeftIndent': 0,
                'RightIndent': 0
            }
        }

        point_size = style_map.get(style, style_map['default'])['PointSize']
        leading = style_map.get(style, style_map['default'])['Leading']
        left_indent = style_map.get(style, style_map['default'])['LeftIndent']
        right_indent = style_map.get(style, style_map['default'])['RightIndent']

        length = story.length
        inner_width -= float(left_indent) + float(right_indent)
        length *= float(point_size) / style_map['default']['PointSize']
        length *= float(leading) / style_map['default']['Leading']
        height = length / inner_width * 75 + 10

        return height



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging

from eve.utils import config
from flask import g, current_app as app

import superdesk
from superdesk import get_resource_service
from superdesk.activity import ACTIVITY_CREATE, ACTIVITY_EVENT, ACTIVITY_UPDATE, notify_and_add_activity, \
    ACTIVITY_DELETE
from superdesk.errors import SuperdeskApiError
from superdesk.io import allowed_feeding_services, allowed_feed_parsers, registered_feeding_services
from superdesk.metadata.item import CONTENT_STATE, content_type
from superdesk.notification import push_notification
from superdesk.resource import Resource
from superdesk.services import BaseService
from superdesk.utc import utcnow
from superdesk.utils import required_string

logger = logging.getLogger(__name__)


class IngestProviderResource(Resource):
    """Ingest provider model

    :param name: provider name
    :param source: populates item source field
    :param feeding_service: feeding service name
    :param feed_parser: feed parser name
    :param content_types: list of content types of items to ingest from provider
    :param allow_remove_ingested: allow deleting of items from ingest
    :param content_expiry: ttl for ingested items in minutes
    :param config: provider specific config
    :param private: can contain any data useful for provider (e.g. to manage feeds position)
    :param ingested_count: number of items ingested so far
    :param tokens: auth tokens used by provider
    :param is_closed: provider closed status
    :param update_schedule: update schedule, will run every x hours x minutes x seconds
    :param idle_time: usual idle time for provider, if there is no item after that it will warn
    :param last_updated: last update timestamp
    :param rule_set: rule sets used when ingesting
    :param routing_scheme: routing scheme used when ingesting
    :param notifications: set when notification should be sent for this provider
    :param last_closed: info when and by whom provider was closed last time
    :param last_opened: info when and by whom provider was opened last time
    :param critical_errors: error codes which are considered critical and should close provider
    """

    def __init__(self, endpoint_name, app, service, endpoint_schema=None):
        self.schema = {
            'name': {
                'type': 'string',
                'required': True,
                'nullable': False,
                'empty': False,
                'iunique': True
            },
            'source': required_string,
            'feeding_service': {
                'type': 'string',
                'required': True,
                'allowed': allowed_feeding_services
            },
            'feed_parser': {
                'type': 'string',
                'nullable': True,
                'allowed': allowed_feed_parsers
            },
            'content_types': {
                'type': 'list',
                'default': content_type,
                'allowed': content_type
            },
            'allow_remove_ingested': {
                'type': 'boolean',
                'default': False
            },
            'content_expiry': {
                'type': 'integer',
                'default': app.config['INGEST_EXPIRY_MINUTES']
            },
            'config': {
                'type': 'dict'
            },
            'private': {
                'type': 'dict'
            },
            'ingested_count': {
                'type': 'integer'
            },
            'accepted_count': {
                'type': 'integer'
            },
            'tokens': {
                'type': 'dict'
            },
            'is_closed': {
                'type': 'boolean',
                'default': False
            },
            'update_schedule': {
                'type': 'dict',
                'schema': {
                    'hours': {'type': 'integer'},
                    'minutes': {'type': 'integer', 'default': 5},
                    'seconds': {'type': 'integer'},
                }
            },
            'idle_time': {
                'type': 'dict',
                'schema': {
                    'hours': {'type': 'integer'},
                    'minutes': {'type': 'integer'},
                }
            },
            'last_updated': {'type': 'datetime'},
            'last_ingested_id': {'type': 'string'},  # this id is ingest provider internal value
            'last_item_update': {'type': 'datetime'},
            'rule_set': Resource.rel('rule_sets', nullable=True),
            'routing_scheme': Resource.rel('routing_schemes', nullable=True),
            'notifications': {
                'type': 'dict',
                'schema': {
                    'on_update': {'type': 'boolean', 'default': True},
                    'on_close': {'type': 'boolean', 'default': True},
                    'on_open': {'type': 'boolean', 'default': True},
                    'on_error': {'type': 'boolean', 'default': True}
                }
            },
            'last_closed': {
                'type': 'dict',
                'schema': {
                    'closed_at': {'type': 'datetime'},
                    'closed_by': Resource.rel('users', nullable=True),
                    'message': {'type': 'string'}
                }
            },
            'last_opened': {
                'type': 'dict',
                'schema': {
                    'opened_at': {'type': 'datetime'},
                    'opened_by': Resource.rel('users', nullable=True)
                }
            },
            'critical_errors': {
                'type': 'dict',
                'valueschema': {
                    'type': 'boolean'
                }
            },
        }

        self.item_methods = ['GET', 'PATCH', 'DELETE']
        self.privileges = {'POST': 'ingest_providers', 'PATCH': 'ingest_providers', 'DELETE': 'ingest_providers'}
        self.etag_ignore_fields = ['last_updated', 'last_item_update', 'last_closed', 'last_opened']

        super().__init__(endpoint_name, app, service, endpoint_schema=endpoint_schema)


class IngestProviderService(BaseService):
    def __init__(self, datasource=None, backend=None):
        super().__init__(datasource=datasource, backend=backend)

    def _set_provider_status(self, doc, message=''):
        user = getattr(g, 'user', None)
        if doc.get('is_closed', True):
            doc['last_closed'] = doc.get('last_closed', {})
            doc['last_closed']['closed_at'] = utcnow()
            doc['last_closed']['closed_by'] = user['_id'] if user else None
            doc['last_closed']['message'] = message
        else:
            doc['last_opened'] = doc.get('last_opened', {})
            doc['last_opened']['opened_at'] = utcnow()
            doc['last_opened']['opened_by'] = user['_id'] if user else None

    @property
    def user_service(self):
        return get_resource_service('users')

    def on_create(self, docs):
        for doc in docs:
            content_expiry = doc.get('content_expiry', 0)
            if content_expiry == 0:
                doc['content_expiry'] = app.config['INGEST_EXPIRY_MINUTES']
                self._set_provider_status(doc, doc.get('last_closed', {}).get('message', ''))
            elif content_expiry < 0:
                doc['content_expiry'] = None
            self._test_config(doc)

    def on_created(self, docs):
        for doc in docs:
            notify_and_add_activity(ACTIVITY_CREATE, 'Created Ingest Channel {{name}}',
                                    self.datasource, item=None,
                                    user_list=self.user_service.get_users_by_user_type('administrator'),
                                    name=doc.get('name'), provider_id=doc.get('_id'))
            push_notification('ingest_provider:create', provider_id=str(doc.get('_id')))
        logger.info("Created Ingest Channel. Data:{}".format(docs))

    def on_update(self, updates, original):
        try:
            content_expiry = updates['content_expiry']
        except KeyError:
            content_expiry = None
        else:
            if content_expiry == 0:
                content_expiry = app.config['INGEST_EXPIRY_MINUTES']
            elif content_expiry < 0:
                content_expiry = None
            updates['content_expiry'] = content_expiry
        if 'is_closed' in updates and original.get('is_closed', False) != updates.get('is_closed'):
            self._set_provider_status(updates, updates.get('last_closed', {}).get('message', ''))
        if 'config' in updates:
            self._test_config(updates, original)

    def on_updated(self, updates, original):
        do_notification = updates.get('notifications', {})\
            .get('on_update', original.get('notifications', {}).get('on_update', True))
        notify_and_add_activity(ACTIVITY_UPDATE, 'updated Ingest Channel {{name}}',
                                self.datasource, item=None,
                                user_list=self.user_service.get_users_by_user_type('administrator')
                                if do_notification else None,
                                name=updates.get('name', original.get('name')),
                                provider_id=original.get('_id'))

        if updates.get('is_closed', False) != original.get('is_closed', False):
            status = ''
            do_notification = False

            if updates.get('is_closed'):
                status = 'closed'
                do_notification = updates.get('notifications', {}). \
                    get('on_close', original.get('notifications', {}).get('on_close', True))
            elif not updates.get('is_closed'):
                status = 'opened'
                do_notification = updates.get('notifications', {}). \
                    get('on_open', original.get('notifications', {}).get('on_open', True))

            notify_and_add_activity(ACTIVITY_EVENT, '{{status}} Ingest Channel {{name}}',
                                    self.datasource, item=None,
                                    user_list=self.user_service.get_users_by_user_type('administrator')
                                    if do_notification else None,
                                    name=updates.get('name', original.get('name')),
                                    status=status, provider_id=original.get('_id'))

        push_notification('ingest_provider:update', provider_id=str(original.get('_id')))
        logger.info("Updated Ingest Channel. Data: {}".format(updates))

    def on_delete(self, doc):
        """
        Overriding to check if the Ingest Source which has received item being deleted.
        """

        if doc.get('last_item_update'):
            raise SuperdeskApiError.forbiddenError("Deleting an Ingest Source after receiving items is prohibited.")

    def on_deleted(self, doc):
        """
        Overriding to send notification and record activity about channel deletion.
        """
        notify_and_add_activity(ACTIVITY_DELETE, 'Deleted Ingest Channel {{name}}',
                                self.datasource, item=None,
                                user_list=self.user_service.get_users_by_user_type('administrator'),
                                name=doc.get('name'), provider_id=doc.get(config.ID_FIELD))
        push_notification('ingest_provider:delete', provider_id=str(doc.get(config.ID_FIELD)))
        get_resource_service('sequences').delete(lookup={
            'key': 'ingest_providers_{_id}'.format(_id=doc[config.ID_FIELD])
        })
        logger.info("Deleted Ingest Channel. Data:{}".format(doc))

    def _test_config(self, updates, original=None):
        provider = original.copy() if original else {}
        provider.update(updates)

        try:
            service = registered_feeding_services[provider['feeding_service']].__class__()
        except KeyError:
            return
        service.config_test(provider)


superdesk.workflow_state(CONTENT_STATE.INGESTED)
superdesk.workflow_action(name='ingest')



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import os
import bcrypt
import hashlib
import base64
import tempfile

from uuid import uuid4
from datetime import datetime
from bson import ObjectId
from enum import Enum
from importlib import import_module
from eve.utils import config
from superdesk.default_settings import ELASTIC_DATE_FORMAT, \
    ELASTIC_DATETIME_FORMAT
from superdesk.text_utils import get_text


required_string = {'type': 'string', 'required': True, 'nullable': False, 'empty': False}


class FileSortAttributes(Enum):
    """
    Enum defining the File Story Attributes.
    """

    name = 1
    created = 2
    modified = 3


class SortOrder(Enum):
    """
    Enum defining the sort order.
    """

    asc = 1
    desc = 2


class SuperdeskBaseEnum(Enum):
    """
    Base enum
    """

    @classmethod
    def from_value(cls, value):
        """
        Returns the valid enum if value found else none

        :param value: enum value
        :return: return valid
        """
        try:
            return cls(value)
        except ValueError:
            return None

    @classmethod
    def values(cls):
        """
        Returns list of values for an enum

        :return: list of values for an enum
        """
        return [enum_member.value for enum_member in cls]


def get_random_string(length=12):
    return str(uuid4())


def get_random_token(n=40):
    """Generate random token.

    :param n: how many random bytes to generate
    """
    return base64.b64encode(os.urandom(n)).decode()


def import_by_path(path):
    module_path, class_name = path.rsplit('.', 1)
    module = import_module(module_path)
    return getattr(module, class_name)


def get_hash(input_str, salt):
    hashed = bcrypt.hashpw(input_str.encode('UTF-8'), bcrypt.gensalt(salt))
    return hashed.decode('UTF-8')


def get_sorted_files(path, sort_by=FileSortAttributes.name, sort_order=SortOrder.asc):
    """
    Get the list of files based on the sort order.

    Sort is allowed on name, created and modified datetime

    :param path: directory path
    :param sort_by: "name", "created", "modified"
    :param sort_order: "asc" - ascending, "desc" - descending
    :return: list of files from the path
    """
    # get the files
    files = [file for file in os.listdir(path) if os.path.isfile(os.path.join(path, file))]
    if sort_by == FileSortAttributes.name:
        files.sort(reverse=(sort_order == SortOrder.desc))
    elif sort_by == FileSortAttributes.created:
        files.sort(key=lambda file: os.path.getctime(os.path.join(path, file)), reverse=(sort_order == SortOrder.desc))
    elif sort_by == FileSortAttributes.modified:
        files.sort(key=lambda file: os.path.getmtime(os.path.join(path, file)), reverse=(sort_order == SortOrder.desc))
    else:
        files.sort(reverse=(sort_order == SortOrder.desc))

    return files


def is_hashed(input_str):
    """Check if given input_str is hashed."""
    return input_str.startswith('$2')


def merge_dicts(dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict, precedence goes to latter dicts.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def merge_dicts_deep(dict1, dict2):
    """
    Deep merge of two dictionaries.
    Example: merged_dict = dict(merge_dicts_deep(dict1, dict2))

    :param dict1: first dictionary
    :param dict2: second dictionary
    :return: generator which will build a merged dict
    """
    unique_keys = set(
        dict1.keys()
    ).union(
        dict2.keys()
    )

    for k in unique_keys:
        # need to merge
        if k in dict1 and k in dict2:
            if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
                yield (k, dict(merge_dicts_deep(dict1[k], dict2[k])))
            else:
                # if one of the values is not a dict - value from second dict overrides value from first one.
                yield (k, dict2[k])
        elif k in dict1:
            yield (k, dict1[k])
        else:
            yield (k, dict2[k])


class ListCursor(object):
    """Wrapper for a python list as a cursor."""

    def __init__(self, docs=None):
        self.docs = docs if docs else []

    def __getitem__(self, key):
        return self.docs[key]

    def first(self):
        """Get first doc."""
        return self.docs[0] if self.docs else None

    def count(self, **kwargs):
        """Get total count."""
        return len(self.docs)

    def extra(self, response):
        pass


def json_serialize_datetime_objectId(obj):
    """
    Serialize so that objectid and date are converted to appropriate format.
    """
    if isinstance(obj, datetime):
        return str(datetime.strftime(obj, config.DATE_FORMAT))

    if isinstance(obj, ObjectId):
        return str(obj)


def compare_preferences(original, updates):
    original_keys = set(original.keys())
    updates_keys = set(updates.keys())
    intersect_keys = original_keys.intersection(updates_keys)
    added = updates_keys - original_keys
    removed = original_keys - updates_keys
    modified = {o: (original[o], updates[o]) for o in intersect_keys if original[o] != updates[o]}
    return added, removed, modified


def sha(text):
    """Get sha hext digest for given text.

    Using sha256 hashing function, returning 64 hex characters.

    :param text: text str
    """
    return hashlib.sha256(text.encode()).hexdigest()


def plaintext_filter(value):
    """Filter out html from value."""
    return get_text(value).replace('\n', ' ').strip()


def format_date(date_string):
    return datetime.strftime(date_string, ELASTIC_DATE_FORMAT)


def format_time(datetime_string):
    return datetime.strftime(datetime_string, ELASTIC_DATETIME_FORMAT)


def save_error_data(data, prefix='superdesk-', suffix='.txt'):
    """Save given data into file and return its name.

    :param data: unicode data
    """
    with tempfile.NamedTemporaryFile(prefix=prefix, suffix=suffix, delete=False) as file:
        file.write(data.encode('utf-8'))
        return file.name



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


from superdesk.services import Service
from superdesk.notification import push_notification
from eve.utils import config
from eve.utils import ParsedRequest


class ContactsService(Service):

    def get(self, req, lookup):

        # by default the response will have the inactive and not public entries filtered out
        if 'all' not in req.args:
            lookup['is_active'] = True
            lookup['public'] = True
        return super().get(req, lookup)

    def on_created(self, docs):
        """
        Send notification to clients that new contact(s) have been created
        :param docs:
        :return:
        """
        push_notification('contacts:create', _id=[doc.get(config.ID_FIELD) for doc in docs])

    def on_updated(self, updates, original):
        """
        Send notifification to clients that a contact has been updated
        :param updates:
        :param original:
        :return:
        """
        push_notification('contacts:update', _id=[original.get(config.ID_FIELD)])

    def on_deleted(self, doc):
        """
        Send a notification to clients that a contact has been deleted
        :param doc:
        :return:
        """
        push_notification('contacts:deleted', _id=[doc.get(config.ID_FIELD)])


class OrganisationService(Service):
    def get(self, req, lookup):
        """
        Search for organisation matching the passed q parameter
        :param req:
        :param lookup:
        :return: List of matching organisation strings
        """
        new_req = ParsedRequest()

        q_str = 'organisation:' + '* organisation:'.join(req.args.get('q', '').split()) + '*'
        new_req.args = {'q': q_str, 'default_operator': 'AND', 'projections': '{"organisation": 1}'}
        ret = super().get(new_req, lookup)

        # Remove any duplicate entries from the response
        orgs = []
        de_duped = []
        for d in ret.docs:
            if d.get('organisation') not in orgs:
                orgs.append(d.get('organisation'))
                de_duped.append(d)
        ret.docs = de_duped
        ret.hits['hits']['total'] = len(ret.docs)
        return ret




import arrow
import superdesk

from arrow.parser import ParserError
from flask import json
from bson import ObjectId
from bson.errors import InvalidId
from eve.utils import str_to_date
from eve.io.mongo import MongoJSONEncoder
from eve_elastic import ElasticJSONSerializer


class SuperdeskJSONEncoder(MongoJSONEncoder, ElasticJSONSerializer):
    """Custom JSON encoder for elastic that can handle `bson.ObjectId`s."""

    pass


def try_cast(v):
    """Cast string value to date or ObjectId if possible.

    :param v: string value
    """
    try:
        str_to_date(v)  # try if it matches format
        return arrow.get(v).datetime  # return timezone aware time
    except (ValueError, ParserError):
        try:
            return ObjectId(v)
        except InvalidId:
            return v


def cast_item(o):
    with superdesk.app.app_context():
        if isinstance(o, (int, float, bool)):
            return o
        elif isinstance(o, str):
            return try_cast(o)
        elif isinstance(o, list):
            for i, v in enumerate(o):
                o[i] = cast_item(v)
            return o
        elif isinstance(o, dict):
            for k, v in o.items():
                o[k] = cast_item(v)
            return o
        else:
            return o


def loads(s):
    o = json.loads(s)

    if isinstance(o, list):
        for i, v in enumerate(o):
            o[i] = cast_item(v)
        return o
    elif isinstance(o, dict):
        for k, v in o.items():
            o[k] = cast_item(v)
        return o
    else:
        return cast_item(o)


def dumps(o):
    with superdesk.app.app_context():
        return MongoJSONEncoder().encode(o)




import superdesk


def init_app(app):
    superdesk.register_default_user_preference('monitoring:view', {
        'type': 'string',
        'allowed': ['list', 'swimlane'],
        'view': 'list',
        'default': 'list',
        'label': 'Monitoring view',
        'category': 'monitoring',
    })

    superdesk.register_default_session_preference('monitoring:view:session', None)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import re
import superdesk
from flask import Blueprint, current_app as app
from datetime import datetime
from eve.render import send_response


NEWSCODE_PATTERN = re.compile('[0-9]{8,9}')
bp = Blueprint('subjectcodes', __name__)


class SubjectIndex():
    """Subjects index."""

    def __init__(self):
        self.subjects = {}
        self.last_modified = datetime.fromtimestamp(0)

    def register(self, subjects, last_modified=None):
        """Register subjects.

        :param subjects: dict with subject qcode as key and name as value
        :param last_modified: datetime for last subjects modification, last one is sent to clients
        """
        self.subjects.update(subjects)
        if last_modified is not None:
            self.last_modified = max(self.last_modified, last_modified)

    def get_items(self):
        """Get list of all subjects.

        Each subject is a dict with `qcode`, `name` and `parent` keys.
        """
        items = []
        for code in sorted(self.subjects):
            items.append({'qcode': code, 'name': self.subjects[code], 'parent': get_parent_subjectcode(code)})
        return items


def get_parent_subjectcode(code):
    """Compute parent code for iptc newscode.

    :param code: iptc newscode without `subj:` prefix
    """
    parent_code = None
    if not NEWSCODE_PATTERN.match(code):
        return parent_code
    if code[-3:] != '000':
        parent_code = code[:5] + '000'
    elif code[2:5] != '000':
        parent_code = code[:2] + '000000'
    return parent_code


@bp.route('/subjectcodes/', methods=['GET', 'OPTIONS'])
def render_subjectcodes():
    items = get_subjectcodeitems()
    response_data = {'_items': items, '_meta': {'total': len(items)}}
    return send_response(None, (response_data, app.subjects.last_modified, None, 200))


def get_subjectcodeitems():
    """Get subjects for current app."""
    return app.subjects.get_items()


def init_app(app):
    app.subjects = SubjectIndex()
    superdesk.blueprint(bp, app)



import logging
import requests
import superdesk

from urllib.parse import urljoin
from flask import current_app as app
from superdesk.utils import ListCursor

logger = logging.getLogger(__name__)


def format_geoname_item(item):
    return {
        'scheme': 'geonames',
        'code': str(item['geonameId']),
        'name': item['name'],
        'state': item.get('adminName1'),
        'region': item.get('adminName2'),
        'country': item.get('countryName'),
        'state_code': item.get('adminCode1'),
        'region_code': item.get('adminCode2'),
        'country_code': item.get('countryCode'),
        'continent_code': item.get('continentCode'),
        'feature_class': item.get('fcl'),
        'location': {
            'lat': float(item['lat']),
            'lon': float(item['lng']),
        }
    }


def geonames_request(service, service_params):
    params = [
        ('type', 'json'),
        ('username', app.config.get('GEONAMES_USERNAME', '')),
    ]
    params.extend(service_params)
    url = urljoin(app.config.get('GEONAMES_URL', 'http://api.geonames.org/'), service)
    res = requests.get(url, params, timeout=10)
    if res.status_code != 200:
        res.raise_for_status()
    return res.json()


class PlacesAutocompleteResource(superdesk.Resource):

    resource_methods = ['GET']
    item_methods = []


class PlacesAutocompleteService(superdesk.Service):
    def get(self, req, lookup):
        assert req.args.get('name'), {'name': 1}
        params = [
            ('name_startsWith', req.args.get('name')),
            ('lang', req.args.get('lang')),
            ('featureClass', 'A'),
            ('featureClass', 'P'),
        ]

        json_data = geonames_request('search', params)
        data = [format_geoname_item(item) for item in json_data.get('geonames', [])]
        return ListCursor(data)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging
import superdesk

import pytz
from pytz import all_timezones_set

from enum import Enum
from datetime import datetime, timedelta
from superdesk import get_resource_service
from superdesk.resource import Resource
from superdesk.services import BaseService
from superdesk.errors import SuperdeskApiError
from eve.utils import config
from superdesk.metadata.item import CONTENT_STATE
from superdesk.utc import set_time

logger = logging.getLogger(__name__)


class Weekdays(Enum):
    """Weekdays names we use for scheduling."""

    MON = 0
    TUE = 1
    WED = 2
    THU = 3
    FRI = 4
    SAT = 5
    SUN = 6

    @classmethod
    def is_valid_schedule(cls, list_of_days):
        """Test if all days in list_of_days are valid day names.

        :param list list_of_days eg. ['mon', 'tue', 'fri']
        """
        return all([day.upper() in cls.__members__ for day in list_of_days])

    @classmethod
    def is_scheduled_day(cls, today, list_of_days):
        """Test if today's weekday is in schedule.

        :param datetime today
        :param list list_of_days
        """
        return today.weekday() in [cls[day.upper()].value for day in list_of_days]

    @classmethod
    def dayname(cls, day):
        """Get name shortcut (MON, TUE, ...) for given day.

        :param datetime day
        """
        return cls(day.weekday()).name


class RoutingRuleSchemeResource(Resource):
    """
    Resource class for 'routing_schemes' endpoint
    """

    schema = {
        'name': {
            'type': 'string',
            'iunique': True,
            'required': True,
            'nullable': False,
            'empty': False
        },
        'rules': {
            'type': 'list',
            'schema': {
                'type': 'dict',
                'schema': {
                    'name': {
                        'type': 'string'
                    },
                    'filter': Resource.rel('content_filters', nullable=True),
                    'actions': {
                        'type': 'dict',
                        'schema': {
                            'fetch': {
                                'type': 'list',
                                'schema': {
                                    'type': 'dict',
                                    'schema': {
                                        'desk': Resource.rel('desks', True),
                                        'stage': Resource.rel('stages', True),
                                        'macro': {'type': 'string'}
                                    }
                                }
                            },
                            'publish': {
                                'type': 'list',
                                'schema': {
                                    'type': 'dict',
                                    'schema': {
                                        'desk': Resource.rel('desks', True),
                                        'stage': Resource.rel('stages', True),
                                        'macro': {'type': 'string'},
                                        'target_subscribers': {
                                            'type': 'list',
                                            'nullable': True
                                        },
                                        'target_types': {
                                            'type': 'list',
                                            'nullable': True
                                        }
                                    }
                                }
                            },
                            'exit': {
                                'type': 'boolean'
                            },
                            'preserve_desk': {
                                'type': 'boolean'
                            }
                        }
                    },
                    'schedule': {
                        'type': 'dict',
                        'nullable': True,
                        'schema': {
                            'day_of_week': {
                                'type': 'list'
                            },
                            'hour_of_day_from': {
                                'type': 'string',
                                'nullable': True,
                            },
                            'hour_of_day_to': {
                                'type': 'string',
                                'nullable': True,
                            },
                            'time_zone': {
                                'type': 'string',
                                'nullable': False,
                                'default': 'UTC'
                            }
                        }
                    }
                }
            }
        }
    }

    privileges = {'POST': 'routing_rules', 'DELETE': 'routing_rules', 'PATCH': 'routing_rules'}


class RoutingRuleSchemeService(BaseService):
    """
    Service class for 'routing_schemes' endpoint.
    """

    def on_create(self, docs):
        """Overriding to check the below pre-conditions:

            1. A routing scheme must have at least one rule.
            2. Every rule in the routing scheme must have name, filter and at least one action

        Will throw BadRequestError if any of the pre-conditions fail.
        """
        for routing_scheme in docs:
            self._adjust_for_empty_schedules(routing_scheme)
            self._validate_routing_scheme(routing_scheme)
            self._check_if_rule_name_is_unique(routing_scheme)

    def on_update(self, updates, original):
        """Overriding to check the below pre-conditions:

            1. A routing scheme must have at least one rule.
            2. Every rule in the routing scheme must have name, filter and at least one action

        Will throw BadRequestError if any of the pre-conditions fail.
        """
        self._adjust_for_empty_schedules(updates)
        self._validate_routing_scheme(updates)
        self._check_if_rule_name_is_unique(updates)

    def on_delete(self, doc):
        """Overriding to check the below pre-conditions:

            1. A routing scheme shouldn't be associated with an Ingest Provider.

        Will throw BadRequestError if any of the pre-conditions fail.
        """

        if self.backend.find_one('ingest_providers', req=None, routing_scheme=doc[config.ID_FIELD]):
            raise SuperdeskApiError.forbiddenError('Routing scheme is applied to channel(s). It cannot be deleted.')

    def apply_routing_scheme(self, ingest_item, provider, routing_scheme):
        """Applies routing scheme and applies appropriate action (fetch, publish) to the item

        :param item: ingest item to which routing scheme needs to applied.
        :param provider: provider for which the routing scheme is applied.
        :param routing_scheme: routing scheme.
        """
        rules = routing_scheme.get('rules', [])
        if not rules:
            logger.warning("Routing Scheme % for provider % has no rules configured." %
                           (provider.get('name'), routing_scheme.get('name')))

        filters_service = superdesk.get_resource_service('content_filters')

        now = datetime.utcnow()

        for rule in self._get_scheduled_routing_rules(rules, now):
            content_filter = rule.get('filter', {})
            logger.info('Applying rule. Item: %s . Routing Scheme: %s. Rule Name %s.' % (ingest_item.get('guid'),
                                                                                         routing_scheme.get('name'),
                                                                                         rule.get('name')))
            if filters_service.does_match(content_filter, ingest_item):
                logger.info('Filter matched. Item: %s. Routing Scheme: %s. Rule Name %s.' % (ingest_item.get('guid'),
                                                                                             routing_scheme.get('name'),
                                                                                             rule.get('name')))
                if rule.get('actions', {}).get('preserve_desk', False) and ingest_item.get('task', {}).get('desk'):
                    desk = get_resource_service('desks').find_one(req=None, _id=ingest_item['task']['desk'])
                    if ingest_item.get('task', {}).get('stage'):
                        stage_id = ingest_item['task']['stage']
                    else:
                        stage_id = desk['incoming_stage']
                    self.__fetch(ingest_item, [{'desk': desk[config.ID_FIELD], 'stage': stage_id}])
                    fetch_actions = [f for f in rule.get('actions', {}).get('fetch', [])
                                     if f.get('desk') != ingest_item['task']['desk']]
                else:
                    fetch_actions = rule.get('actions', {}).get('fetch', [])

                self.__fetch(ingest_item, fetch_actions)
                self.__publish(ingest_item, rule.get('actions', {}).get('publish', []))
                if rule.get('actions', {}).get('exit', False):
                    logger.info('Exiting routing scheme. Item: %s . Routing Scheme: %s. '
                                'Rule Name %s.' % (ingest_item.get('guid'), routing_scheme.get('name'),
                                                   rule.get('name')))
                    break
            else:
                logger.info("Routing rule %s of Routing Scheme %s for Provider %s did not match for item %s" %
                            (rule.get('name'), routing_scheme.get('name'),
                             provider.get('name'), ingest_item[config.ID_FIELD]))

    def _adjust_for_empty_schedules(self, routing_scheme):
        """Adjust for empty schedules.

        For all routing scheme's rules, set their non-empty schedules to
        None if they are effectively not defined.

        A schedule is recognized as "not defined" if it only contains time zone
        information without anything else. This can happen if an empty schedule
        is submitted by the client, because `Eve` then converts it to the
        following:

            {'time_zone': 'UTC'}

        This is because the time_zone field has a default value set in the
        schema, and Eve wants to apply it even when the containing object (i.e.
        the schedule) is None and there is nothing that would contain the time
        zone information.

        :param dict routing_scheme: the routing scheme to check
        """
        for rule in routing_scheme.get('rules', []):
            schedule = rule.get('schedule')
            if schedule:
                if set(schedule.keys()) == {'time_zone'}:
                    rule['schedule'] = None
                elif 'time_zone' not in schedule.keys():
                    schedule['time_zone'] = 'UTC'

    def _validate_routing_scheme(self, routing_scheme):
        """Validates routing scheme for the below:

            1. A routing scheme must have at least one rule.
            2. Every rule in the routing scheme must have name, filter and at least one action

        Will throw BadRequestError if any of the conditions fail.

        :param routing_scheme:
        """

        routing_rules = routing_scheme.get('rules', [])
        if len(routing_rules) == 0:
            raise SuperdeskApiError.badRequestError(message="A Routing Scheme must have at least one Rule")
        for routing_rule in routing_rules:
            invalid_fields = [field for field in routing_rule.keys()
                              if field not in ('name', 'filter', 'actions', 'schedule')]

            if invalid_fields:
                raise SuperdeskApiError.badRequestError(
                    message="A routing rule has invalid fields %s".format(invalid_fields))

            schedule = routing_rule.get('schedule')
            actions = routing_rule.get('actions')

            if routing_rule.get('name') is None:
                raise SuperdeskApiError.badRequestError(message="A routing rule must have a name")
            elif actions is None or len(actions) == 0 or (actions.get('fetch') is None and actions.get(
                    'publish') is None and actions.get('exit') is None):
                raise SuperdeskApiError.badRequestError(message="A routing rule must have actions")
            else:
                self._validate_schedule(schedule)

    def _validate_schedule(self, schedule):
        """Validate schedule.

        Check if the given routing schedule configuration is valid and raise
        an error if this is not the case.

        :param dict schedule: the routing schedule configuration to validate

        :raises SuperdeskApiError: if validation of `schedule` fails
        """
        if schedule is not None and (
            len(schedule) == 0 or
            schedule.get('day_of_week') is None or
            len(schedule.get('day_of_week', [])) == 0
        ):
            raise SuperdeskApiError.badRequestError(message="Schedule when defined can't be empty.")

        if schedule:
            if not Weekdays.is_valid_schedule(schedule.get('day_of_week', [])):
                raise SuperdeskApiError.badRequestError(message="Invalid values for day of week.")

            if schedule.get('hour_of_day_from') or schedule.get('hour_of_day_to'):
                try:
                    from_time = datetime.strptime(schedule.get('hour_of_day_from'), '%H:%M:%S')
                except Exception:
                    raise SuperdeskApiError.badRequestError(message="Invalid value for from time.")

                to_time = schedule.get('hour_of_day_to', '')
                if to_time:
                    try:
                        to_time = datetime.strptime(to_time, '%H:%M:%S')
                    except Exception:
                        raise SuperdeskApiError.badRequestError(
                            message="Invalid value for hour_of_day_to "
                                    "(expected %H:%M:%S).")

                    if from_time > to_time:
                        raise SuperdeskApiError.badRequestError(
                            message="From time should be less than To Time."
                        )

            time_zone = schedule.get('time_zone')

            if time_zone and (time_zone not in all_timezones_set):
                msg = 'Unknown time zone {}'.format(time_zone)
                raise SuperdeskApiError.badRequestError(message=msg)

    def _check_if_rule_name_is_unique(self, routing_scheme):
        """
        Checks if name of a routing rule is unique or not.
        """
        routing_rules = routing_scheme.get('rules', [])

        for routing_rule in routing_rules:
            rules_with_same_name = [rule for rule in routing_rules if rule.get('name') == routing_rule.get('name')]

            if len(rules_with_same_name) > 1:
                raise SuperdeskApiError.badRequestError("Rule Names must be unique within a scheme")

    def _get_scheduled_routing_rules(self, rules, current_dt_utc):
        """
        Iterates rules list and returns the list of rules that are scheduled.

        :param list rules: routing rules to check
        :param datetime current_dt_utc: the value to take as the current
            time in UTC

        :return: the rules scheduled to be appplied at `current_dt_utc`
        :rtype: list
        """
        # make it a timezone-aware object
        current_dt_utc = current_dt_utc.replace(tzinfo=pytz.utc)
        delta_minute = timedelta(minutes=1)

        scheduled_rules = []
        for rule in rules:
            is_scheduled = True
            schedule = rule.get('schedule', {})
            if schedule:
                # adjust current time to the schedule's timezone
                tz_name = schedule.get('time_zone')
                schedule_tz = pytz.timezone(tz_name) if tz_name else pytz.utc
                now_tz_schedule = current_dt_utc.astimezone(tz=schedule_tz)

                # Create start and end time-of-day limits. If start time is not
                # defined, the beginning of the day is assumed. If end time
                # is not defined, the end of the day is assumed (excluding the
                # midnight, since at that point a new day has already begun).
                hour_of_day_from = schedule.get('hour_of_day_from')
                if not hour_of_day_from:
                    hour_of_day_from = '00:00:00'  # might be both '' or None
                from_time = set_time(now_tz_schedule, hour_of_day_from)

                hour_of_day_to = schedule.get('hour_of_day_to')
                if hour_of_day_to:
                    to_time = set_time(now_tz_schedule, hour_of_day_to)
                    if hour_of_day_to[-2:] == '00':
                        to_time = to_time + delta_minute
                else:
                    to_time = set_time(now_tz_schedule, '23:59:59') + delta_minute

                # check if the current day of week and time of day both match
                day_of_week_matches = Weekdays.is_scheduled_day(
                    now_tz_schedule, schedule.get('day_of_week', []))
                time_of_day_matches = (from_time <= now_tz_schedule < to_time)

                is_scheduled = (day_of_week_matches and time_of_day_matches)

            if is_scheduled:
                scheduled_rules.append(rule)

        return scheduled_rules

    def __fetch(self, ingest_item, destinations):
        """Fetch to item to the destinations

        :param item: item to be fetched
        :param destinations: list of desk and stage
        """
        archive_items = []
        for destination in destinations:
            try:
                logger.info('Fetching item %s to desk %s' % (ingest_item.get('guid'), destination))
                target = self.__getTarget(destination)
                item_id = get_resource_service('fetch') \
                    .fetch([{config.ID_FIELD: ingest_item[config.ID_FIELD],
                             'desk': str(destination.get('desk')),
                             'stage': str(destination.get('stage')),
                             'state': CONTENT_STATE.ROUTED,
                             'macro': destination.get('macro', None),
                             'target': target}])[0]

                archive_items.append(item_id)
                logger.info('Fetched item %s to desk %s' % (ingest_item.get('guid'), destination))
            except Exception:
                logger.exception("Failed to fetch item %s to desk %s" % (ingest_item.get('guid'), destination))

        return archive_items

    def __getTarget(self, destination):
        """Get the target for destination

        :param dict destination: routing destination
        :return dict: returns target information
        """
        target = {}
        if destination.get('target_subscribers'):
            target['target_subscribers'] = destination.get('target_subscribers')

        if destination.get('target_types'):
            target['target_types'] = destination.get('target_types')

        return target

    def __publish(self, ingest_item, destinations):
        """Fetches the item to the desk and then publishes the item.

        :param item: item to be published
        :param destinations: list of desk and stage
        """
        items_to_publish = self.__fetch(ingest_item, destinations)
        for item in items_to_publish:
            try:
                logger.info('Publishing item %s' % (ingest_item.get('guid')))
                self._set_default_values(item)
                get_resource_service('archive_publish').patch(item, {'auto_publish': True})
                logger.info('Published item %s' % (ingest_item.get('guid')))
            except Exception:
                logger.exception("Failed to publish item %s." % item)

    def _set_default_values(self, item):
        """Assigns the default values to the item that about to be auto published"""

        archive_item = get_resource_service('archive').find_one(req=None, _id=item)
        default_values = {}
        default_values['headline'] = archive_item.get('headline') or ' '
        default_values['anpa_category'] = \
            archive_item.get('anpa_category') or [{'qcode': 'a', 'name': 'Australian General News'}]
        default_values['slugline'] = archive_item.get('slugline') or ' '
        default_values['body_html'] = archive_item.get('body_html') or '<p></p>'
        get_resource_service('archive').patch(item, default_values)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from .media_references import MediaReferencesResource
from .media_editor import MediaEditorService, MediaEditorResource
from superdesk.services import BaseService
import superdesk


def init_app(app):
    endpoint_name = 'media_references'
    service = BaseService(endpoint_name, backend=superdesk.get_backend())
    MediaReferencesResource(endpoint_name, app=app, service=service)

    endpoint_name = 'media_editor'
    service = MediaEditorService(endpoint_name, backend=superdesk.get_backend())
    MediaEditorResource(endpoint_name, app=app, service=service)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from superdesk.errors import SuperdeskApiError
import logging
from flask import json


logger = logging.getLogger(__name__)


class UserDisabledError(SuperdeskApiError):
    """User is disabled, access restricted

    """

    status_code = 403
    payload = {'is_enabled': False}
    message = 'Account is disabled, access restricted.'


class PasswordExpiredError(SuperdeskApiError):
    """The password of the user has expired

    """

    status_code = 403
    payload = {'password_is_expired': True}
    message = 'The password of the user has expired.'


class CredentialsAuthError(SuperdeskApiError):
    """Credentials Not Match Auth Exception

    """

    def __init__(self, credentials, message=None, error=None):
        super().__init__(status_code=401, message=message, payload={'credentials': 1})
        # pop the password so that it doesn't get logged
        credentials.pop('password', None)
        logger.warning("Login failure: %s" % json.dumps(credentials))
        if error:
            logger.error("Exception occurred: {}".format(error))



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging
import json

from bson.objectid import ObjectId

import superdesk
from copy import deepcopy
from flask import current_app as app
from eve.utils import ParsedRequest
from eve.versioning import versioned_id_field
from superdesk.celery_app import celery
from superdesk import get_resource_service, config
from superdesk.celery_task_utils import get_lock_id
from .resource import LEGAL_ARCHIVE_NAME, LEGAL_ARCHIVE_VERSIONS_NAME, LEGAL_PUBLISH_QUEUE_NAME, \
    LEGAL_ARCHIVE_HISTORY_NAME
from superdesk.users.services import get_display_name
from apps.archive.common import ARCHIVE
from superdesk.metadata.item import ITEM_STATE, CONTENT_STATE
from superdesk.lock import lock, unlock
from superdesk.publish.publish_queue import QueueState
from superdesk.errors import update_notifiers
from superdesk.activity import ACTIVITY_ERROR
from superdesk.utc import utcnow

logger = logging.getLogger(__name__)


def is_legal_archive_enabled():
    """Test if legal archive is enabled."""
    return app.config['LEGAL_ARCHIVE']


class LegalArchiveImport:
    log_msg_format = "{{'_id': {_id}, 'unique_name': {unique_name}, 'version': {_current_version}, " \
                     "'expired_on': {expiry}}}."

    def upsert_into_legal_archive(self, item_id):
        """Once publish actions are performed on the article do the below:

            1.  Get legal archive article.
            2.  De-normalize the expired article
            3.  Upserting Legal Archive.
            4.  Get Version History and De-normalize and Inserting Legal Archive Versions
            5.  Get History and de-normalize and insert into Legal Archive History

        :param dict item_id: id of the document from 'archive' collection.
        """
        try:

            logger.info('Import item into legal {}.'.format(item_id))

            doc = get_resource_service(ARCHIVE).find_one(req=None, _id=item_id)

            if not doc:
                logger.error('Could not find the document {} to import to legal archive.'.format(item_id))
                return

            # setting default values in case they are missing other log message will fail.
            doc.setdefault('unique_name', 'NO UNIQUE NAME')
            doc.setdefault(config.VERSION, 1)
            doc.setdefault('expiry', utcnow())

            if not doc.get(ITEM_STATE) in \
                    {CONTENT_STATE.PUBLISHED, CONTENT_STATE.CORRECTED, CONTENT_STATE.KILLED, CONTENT_STATE.RECALLED}:
                # at times we have seen that item is published but the item is different in the archive collection
                # this will notify admins about the issue but proceed to move the item into legal archive.
                msg = 'Invalid state: {}. Moving the item to legal archive. item: {}'.\
                    format(doc.get(ITEM_STATE), self.log_msg_format.format(**doc))
                logger.error(msg)
                update_notifiers(ACTIVITY_ERROR, msg=msg, resource=ARCHIVE)

            # required for behave test.
            legal_archive_doc = deepcopy(doc)
            legal_archive_service = get_resource_service(LEGAL_ARCHIVE_NAME)
            legal_archive_versions_service = get_resource_service(LEGAL_ARCHIVE_VERSIONS_NAME)
            legal_archive_history_service = get_resource_service(LEGAL_ARCHIVE_HISTORY_NAME)

            log_msg = self.log_msg_format.format(**legal_archive_doc)
            version_id_field = versioned_id_field(app.config['DOMAIN'][ARCHIVE])
            logger.info('Preparing Article to be inserted into Legal Archive ' + log_msg)

            # Removing irrelevant properties
            legal_archive_doc.pop(config.ETAG, None)
            legal_archive_doc.pop('lock_user', None)
            legal_archive_doc.pop('lock_session', None)
            legal_archive_doc.pop('lock_time', None)
            legal_archive_doc.pop('lock_action', None)

            logger.info('Removed irrelevant properties from the article {}'.format(log_msg))

            # Step 1
            article_in_legal_archive = legal_archive_service.find_one(req=None, _id=legal_archive_doc[config.ID_FIELD])

            if article_in_legal_archive and \
               article_in_legal_archive.get(config.VERSION, 0) > legal_archive_doc.get(config.VERSION):
                logger.info('Item {} version: {} already in legal archive. Legal Archive document version {}'.format(
                    legal_archive_doc.get(config.ID_FIELD), legal_archive_doc.get(config.VERSION),
                    article_in_legal_archive.get(config.VERSION)
                ))
                self._set_moved_to_legal(doc)
                return

            # Step 2 - De-normalizing the legal archive doc
            self._denormalize_user_desk(legal_archive_doc, log_msg)
            logger.info('De-normalized article {}'.format(log_msg))

            # Step 3 - Upserting Legal Archive
            logger.info('Upserting Legal Archive Repo with article {}'.format(log_msg))

            if article_in_legal_archive:
                legal_archive_service.put(legal_archive_doc[config.ID_FIELD], legal_archive_doc)
            else:
                legal_archive_service.post([legal_archive_doc])

            # Step 4 - Get Versions and De-normalize and Inserting Legal Archive Versions
            lookup = {version_id_field: legal_archive_doc[config.ID_FIELD]}
            versions = list(get_resource_service('archive_versions').get(req=None, lookup=lookup))
            legal_versions = list(legal_archive_versions_service.get(req=None, lookup=lookup))

            logger.info('Fetched version history for article {}'.format(log_msg))
            versions_to_insert = [version for version in versions
                                  if not any(legal_version for legal_version in legal_versions
                                             if version[config.VERSION] == legal_version[config.VERSION])]

            # Step 5 - Get History and de-normalize and insert into Legal Archive History
            lookup = {'item_id': legal_archive_doc[config.ID_FIELD]}
            history_items = list(get_resource_service('archive_history').get(req=None, lookup=lookup))
            legal_history_items = list(legal_archive_history_service.get(req=None, lookup=lookup))

            logger.info('Fetched history for article {}'.format(log_msg))
            history_to_insert = [history for history in history_items
                                 if not any(legal_version for legal_version in legal_history_items
                                            if history[config.ID_FIELD] == legal_version[config.ID_FIELD])]

            # This happens when user kills an article from Dusty Archive
            if article_in_legal_archive and \
               article_in_legal_archive[config.VERSION] < legal_archive_doc[config.VERSION] and \
               len(versions_to_insert) == 0:

                resource_def = app.config['DOMAIN'][ARCHIVE]
                versioned_doc = deepcopy(legal_archive_doc)
                versioned_doc[versioned_id_field(resource_def)] = legal_archive_doc[config.ID_FIELD]
                versioned_doc[config.ID_FIELD] = ObjectId()
                versions_to_insert.append(versioned_doc)

            for version_doc in versions_to_insert:
                self._denormalize_user_desk(version_doc,
                                            self.log_msg_format.format(_id=version_doc[version_id_field],
                                                                       unique_name=version_doc.get('unique_name'),
                                                                       _current_version=version_doc[config.VERSION],
                                                                       expiry=version_doc.get('expiry')))
                version_doc.pop(config.ETAG, None)

            if versions_to_insert:
                legal_archive_versions_service.post(versions_to_insert)
                logger.info('Inserted de-normalized versions for article {}'.format(log_msg))

            for history_doc in history_to_insert:
                self._denormalize_history(history_doc)
                history_doc.pop(config.ETAG, None)

            if history_to_insert:
                legal_archive_history_service.post(history_to_insert)
                logger.info('Inserted de-normalized history for article {}'.format(log_msg))

            # Set the flag that item is moved to legal.
            self._set_moved_to_legal(doc)

            logger.info('Upsert completed for article ' + log_msg)
        except Exception:
            logger.exception('Failed to import into legal archive {}.'.format(item_id))
            raise

    def _denormalize_history(self, history_item):
        """
        De-normalizes history items
        """
        msg = "item_id: {} and version: {}".format(history_item['item_id'], history_item['version'])

        # De-normalizing User Details
        history_item['user_id'] = self.__get_user_name(history_item.get('user_id'))

        # De-normalizing Desk and Stage details
        history_update = history_item.get('update')
        if history_update:
            if history_update.get('task') and history_update.get('task').get('desk'):
                desk = get_resource_service('desks').find_one(req=None, _id=str(history_update['task']['desk']))
                if desk:
                    history_update['task']['desk'] = desk.get('name')
                    logger.info('De-normalized Desk Details for article history {}'.format(msg))
                else:
                    logger.info('Desk Details Not Found: {}. {}'.format(history_update['task'].get('desk'), msg))

            if history_update.get('task') and history_update['task'].get('stage'):
                stage = get_resource_service('stages').find_one(req=None, _id=str(history_update['task']['stage']))
                if stage:
                    history_update['task']['stage'] = stage.get('name')
                    logger.info('De-normalized Stage Details for article {}'.format(msg))
                else:
                    logger.info('Stage Details Not Found: {}. {}'.format(history_update['task'].get('stage'),
                                                                         msg))

            if history_update.get('task') and history_update['task'].get('user'):
                    history_update['task']['user'] = self.__get_user_name(history_update['task'].get('user'))

            history_item['update'] = history_update

    def _denormalize_user_desk(self, legal_archive_doc, log_msg):
        """
        De-normalizes user, desk and stage details in legal_archive_doc.
        """

        # De-normalizing User Details
        legal_archive_doc['original_creator'] = self.__get_user_name(legal_archive_doc.get('original_creator'))
        legal_archive_doc['version_creator'] = self.__get_user_name(legal_archive_doc.get('version_creator'))

        logger.info('De-normalized User Details for article {}'.format(log_msg))

        # De-normalizing Desk and Stage details
        if legal_archive_doc.get('task'):
            if legal_archive_doc['task'].get('desk'):
                desk = get_resource_service('desks').find_one(req=None, _id=str(legal_archive_doc['task']['desk']))
                if desk:
                    legal_archive_doc['task']['desk'] = desk.get('name')
                    logger.info('De-normalized Desk Details for article {}'.format(log_msg))
                else:
                    logger.info('Desk Details Not Found: {}. {}'.format(legal_archive_doc['task'].get('desk'), log_msg))

            if legal_archive_doc['task'].get('stage'):
                stage = get_resource_service('stages').find_one(req=None, _id=str(legal_archive_doc['task']['stage']))
                if stage:
                    legal_archive_doc['task']['stage'] = stage.get('name')
                    logger.info('De-normalized Stage Details for article {}'.format(log_msg))
                else:
                    logger.info('Stage Details Not Found: {}. {}'.format(legal_archive_doc['task'].get('stage'),
                                                                         log_msg))

            legal_archive_doc['task']['user'] = self.__get_user_name(legal_archive_doc['task'].get('user'))

    def __get_user_name(self, user_id):
        """
        Retrieves display_name of the user identified by user_id
        """
        logger.info('Get User Details for ID:{}'.format(user_id))

        if not user_id:
            return ''

        user = get_resource_service('users').find_one(req=None, _id=user_id)

        if not user:
            return ''

        return get_display_name(user)

    def _set_moved_to_legal(self, doc):
        """Set the moved to legal flag.

        :param dict doc: document
        """
        get_resource_service('published').set_moved_to_legal(doc.get(config.ID_FIELD),
                                                             doc.get(config.VERSION),
                                                             True)

    def import_legal_publish_queue(self, force_move=False, page_size=500):
        """Import legal publish queue.

        :param bool force_move: True force move to legal as else false.
        :param int page_size: page_size.
        """
        logger.info('Starting to import publish queue items...')

        for items in self.get_publish_queue_items(page_size):
            if len(items):
                try:
                    self.process_queue_items(items, force_move)
                    logger.info('Imported publish queue items {} into legal publish queue.'.format(len(items)))
                except Exception:
                    logger.exception('Failed to import into legal publish queue via command')

        logger.info('Completed importing of publish queue items.')

    def process_queue_items(self, queue_items, force_move=False):
        """Process queue items.

        :param list queue_items: list of queue item to be
        :param bool force_move:
        """
        logger.info('Items to import {}.'.format(len(queue_items)))
        logger.info('Get subscribers info for de-normalising queue items.')
        subscriber_ids = list({str(queue_item['subscriber_id']) for queue_item in queue_items})
        query = {'$and': [{config.ID_FIELD: {'$in': subscriber_ids}}]}
        subscribers = list(get_resource_service('subscribers').get(req=None, lookup=query))
        subscribers = {str(subscriber[config.ID_FIELD]): subscriber for subscriber in subscribers}

        for queue_item in queue_items:
            try:
                self._upsert_into_legal_archive_publish_queue(queue_item, subscribers, force_move)
            except Exception:
                logger.exception("Failed to import publish queue item. {}".format(queue_item.get(config.ID_FIELD)))

    def _upsert_into_legal_archive_publish_queue(self, queue_item, subscribers, force_move):
        """Upsert into legal publish queue.

        :param dict queue_item: publish_queue collection item
        :param dict subscribers: subscribers information
        :param bool force_move: true set the flag to move to legal.
        """
        legal_publish_queue_service = get_resource_service(LEGAL_PUBLISH_QUEUE_NAME)
        legal_queue_item = deepcopy(queue_item)
        lookup = {
            'item_id': legal_queue_item.get('item_id'),
            'item_version': legal_queue_item.get('item_version'),
            'subscriber_id': legal_queue_item.get('subscriber_id')
        }

        log_msg = '{item_id} -- version {item_version} -- subscriber {subscriber_id}.'.format(**lookup)

        logger.info('Processing queue item: {}'.format(log_msg))

        existing_queue_item = legal_publish_queue_service.find_one(req=None, _id=legal_queue_item.get(config.ID_FIELD))
        if str(queue_item['subscriber_id']) in subscribers:
            legal_queue_item['subscriber_id'] = subscribers[str(queue_item['subscriber_id'])]['name']
            legal_queue_item['_subscriber_id'] = queue_item['subscriber_id']
        else:
            logger.warn('Subscriber is deleted from the system: {}'.format(log_msg))
            legal_queue_item['subscriber_id'] = 'Deleted Subscriber'
            legal_queue_item['_subscriber_id'] = queue_item['subscriber_id']

        if not existing_queue_item:
            legal_publish_queue_service.post([legal_queue_item])
            logger.info('Inserted queue item: {}'.format(log_msg))
        else:
            legal_publish_queue_service.put(existing_queue_item.get(config.ID_FIELD), legal_queue_item)
            logger.info('Updated queue item: {}'.format(log_msg))

        if queue_item['state'] in {QueueState.SUCCESS.value,
                                   QueueState.CANCELED.value,
                                   QueueState.FAILED.value} or force_move:
            updates = dict()
            updates['moved_to_legal'] = True

            try:
                get_resource_service('publish_queue').system_update(queue_item.get(config.ID_FIELD),
                                                                    updates, queue_item)
                logger.info('Queue item moved to legal. {}'.format(log_msg))
            except Exception:
                logger.exception('Failed to set moved to legal flag for queue item {}.'.format(log_msg))

        logger.info('Processed queue item: {}'.format(log_msg))

    def get_publish_queue_items(self, page_size, expired_items=[]):
        """Get publish queue items that are not moved to legal

        :param int page_size: batch size
        :param list expired_items:
        :return list: publish queue items
        """
        query = {
            'moved_to_legal': False
        }

        if expired_items:
            query['item_id'] = {'$in': expired_items}
        else:
            query['state'] = {'$in': [QueueState.SUCCESS.value, QueueState.CANCELED.value, QueueState.FAILED.value]}

        service = get_resource_service('publish_queue')
        req = ParsedRequest()
        req.sort = '[("_id", 1)]'
        req.where = json.dumps(query)
        cursor = service.get(req=req, lookup=None)
        count = cursor.count()
        no_of_pages = 0
        if count:
            no_of_pages = len(range(0, count, page_size))
            queue_id = cursor[0][config.ID_FIELD]
        logger.info('Number of items to move to legal archive publish queue: {}, pages={}'.format(count, no_of_pages))

        for page in range(0, no_of_pages):
            logger.info('Fetching publish queue items '
                        'for page number: {}. queue_id: {}'. format((page + 1), queue_id))
            req = ParsedRequest()
            req.sort = '[("_id", 1)]'
            query['_id'] = {'$gte': str(queue_id)}
            req.where = json.dumps(query)
            req.max_results = page_size
            cursor = service.get(req=req, lookup=None)
            items = list(cursor)
            if len(items) > 0:
                queue_id = items[len(items) - 1][config.ID_FIELD]
            logger.info('Fetched No. of Items: {} for page: {} '
                        'For import in to legal archive publish_queue.'.format(len(items), (page + 1)))
            yield items


@celery.task(bind=True, default_retry_delay=180)
def import_into_legal_archive(self, item_id):
    """Called async to import into legal archive.

    :param self: celery task
    :param str item_id: document id to import into legal_archive
    """
    if not is_legal_archive_enabled():
        return
    try:
        LegalArchiveImport().upsert_into_legal_archive(item_id)
    except Exception:
        # we can't loose stuff for legal archive.
        logger.exception('Failed to process legal archive doc {}. Retrying again.'.format(item_id))
        raise self.retry()


class ImportLegalPublishQueueCommand(superdesk.Command):
    """
    This command import publish queue records into legal publish queue.
    """

    default_page_size = 500

    option_list = [
        superdesk.Option('--page-size', '-p', dest='page_size', required=False)
    ]

    def run(self, page_size=None):
        if not is_legal_archive_enabled():
            return
        logger.info('Import to Legal Publish Queue')
        lock_name = get_lock_id('legal_archive', 'import_legal_publish_queue')
        page_size = int(page_size) if page_size else self.default_page_size
        if not lock(lock_name, expire=310):
            return
        try:
            LegalArchiveImport().import_legal_publish_queue(page_size=page_size)
        finally:
            unlock(lock_name)


class ImportLegalArchiveCommand(superdesk.Command):
    """This command import archive into legal archive.

    As per the publishing logic the import to legal archive is done asynchronously. If this fails
    then you are missing records in legal archive. Use this command to manually import archive
    items into legal archive.
    """

    default_page_size = 500

    option_list = [
        superdesk.Option('--page-size', '-p', dest='page_size', required=False)
    ]

    def run(self, page_size=None):
        if not is_legal_archive_enabled():
            return
        logger.info('Import to Legal Archive')
        lock_name = get_lock_id('legal_archive', 'import_to_legal_archive')
        page_size = int(page_size) if page_size else self.default_page_size
        if not lock(lock_name, expire=1810):
            return
        try:
            legal_archive_import = LegalArchiveImport()
            # publish_queue = get_resource_service('publish_queue')
            # move the publish item to legal archive.
            expired_items = set()
            for items in self.get_expired_items(page_size):
                for item in items:
                    self._move_to_legal(item.get('item_id'), item.get(config.VERSION), expired_items)

            # get the invalid items from archive.
            for items in get_resource_service(ARCHIVE).get_expired_items(utcnow(), invalid_only=True):
                for item in items:
                    self._move_to_legal(item.get(config.ID_FIELD), item.get(config.VERSION), expired_items)

            # if publish item is moved but publish_queue item is not.
            if len(expired_items):
                try:
                    for items in legal_archive_import.get_publish_queue_items(page_size, list(expired_items)):
                        legal_archive_import.process_queue_items(items, True)
                except Exception:
                    logger.exception('Failed to import into legal publish queue via command')

            # reset the expiry status
            archive_service = get_resource_service(ARCHIVE)
            for item_id in expired_items:
                try:
                    item = archive_service.find_one(req=None, _id=item_id)
                    if item:
                        archive_service.system_update(item_id, {'expiry_status': ''}, item)
                except Exception:
                    logger.exception('Failed to reset expiry status for item id: {}.'.format(item_id))
        except Exception:
            logger.exception('Failed to import into legal archive.')
        finally:
            unlock(lock_name)

    def _move_to_legal(self, item_id, item_version, expired_items):
        try:
            legal_archive_import = LegalArchiveImport()
            legal_archive_import.upsert_into_legal_archive(item_id)
            # set the flag to be set to true.
            get_resource_service('published').set_moved_to_legal(item_id,
                                                                 item_version, True)
            expired_items.add(item_id)
        except Exception:
            logger.exception('Failed to import into legal archive via command {}.'.format(item_id))

    def get_expired_items(self, page_size):
        """Get expired item that are not moved to legal

        :return:
        """
        query = {
            'query': {
                'filtered': {
                    'filter': {
                        'and': [
                            {'range': {'expiry': {'lt': 'now'}}},
                            {'term': {'moved_to_legal': False}},
                            {'not': {'term': {'state': CONTENT_STATE.SCHEDULED}}}
                        ]
                    }
                }
            }
        }

        service = get_resource_service('published')
        req = ParsedRequest()
        req.args = {'source': json.dumps(query)}
        req.sort = '[("publish_sequence_no", 1)]'
        cursor = service.get(req=req, lookup=None)
        count = cursor.count()
        no_of_pages = 0
        if count:
            no_of_pages = len(range(0, count, page_size))
            sequence_no = cursor[0]['publish_sequence_no']
        logger.info('Number of items to move to legal archive: {}, pages={}'.format(count, no_of_pages))

        for page in range(0, no_of_pages):
            logger.info('Fetching published items '
                        'for page number: {} sequence no: {}'. format((page + 1), sequence_no))
            req = ParsedRequest()
            page_query = deepcopy(query)
            sequence_filter = {'range': {'publish_sequence_no': {'gte': sequence_no}}}
            if page == 0:
                sequence_filter = {'range': {'publish_sequence_no': {'gte': sequence_no}}}
            else:
                sequence_filter = {'range': {'publish_sequence_no': {'gt': sequence_no}}}

            page_query['query']['filtered']['filter']['and'].append(sequence_filter)

            req.args = {'source': json.dumps(page_query)}
            req.sort = '[("publish_sequence_no", 1)]'
            req.max_results = page_size
            cursor = service.get(req=req, lookup=None)
            items = list(cursor)
            if len(items):
                sequence_no = items[len(items) - 1]['publish_sequence_no']

            logger.info('Fetched No. of Items: {} for page: {} '
                        'For import into legal archive.'.format(len(items), (page + 1)))
            yield items



from lxml import etree

from .story import Story


class StoryList(Story):
    """
    Story which represents `ul`, `ol` html tags.
    """

    def _add_story(self):
        # merge Story attributes
        story_attributes = self.merge_attributes(
            self.STORY_DEFAULTS,
            self._attributes.get('Story', {})
        )
        story_attributes.update({'Self': self.self_id})
        # Story
        story = etree.SubElement(
            self._etree,
            'Story',
            attrib=story_attributes
        )
        # StoryPreference
        etree.SubElement(
            story,
            'StoryPreference',
            attrib=self.merge_attributes(
                self.STORYPREFERENCE_DEFAULTS,
                self._attributes.get('StoryPreference', {})
            )
        )

        if self._markup_tag:
            # XMLElement to tag a story
            paragraphstylerange_container = etree.SubElement(
                story,
                'XMLElement',
                attrib={
                    'Self': '{}_{}'.format(self.self_id, self._markup_tag.lower()),
                    'XMLContent': self.self_id,
                    'MarkupTag': 'XMLTag/{}'.format(self._markup_tag)
                }
            )
        else:
            paragraphstylerange_container = story

        # ParagraphStyleRange
        paragraphstylerange = etree.SubElement(
            paragraphstylerange_container,
            'ParagraphStyleRange',
            attrib=self.merge_attributes(
                self.PARAGRAPHSTYLERANGE_DEFAULTS,
                self._attributes.get('ParagraphStyleRange', {})
            )
        )

        # CharacterStyleRange(s) + <Br />
        for li in self._element.xpath('.//li'):
            if paragraphstylerange.find('CharacterStyleRange') is not None:
                etree.SubElement(
                    paragraphstylerange,
                    'Br'
                )
            paragraphstylerange[:] += self._handle_inline_tags(li)

        return story

    @property
    def length(self):
        raise NotImplementedError

    @staticmethod
    def guess_height(story, inner_width):
        list_height = 0
        list_item_length = 0
        list_item_height = 0

        for el in story._etree.xpath('.//ParagraphStyleRange')[0].iterchildren():

            if el.tag == 'Br':
                if list_item_height < 20:
                    list_item_height = 20
                list_height += list_item_height
                # next list item
                list_item_length = 0
                list_item_height = 0
            else:
                list_item_length += len(" ".join(etree.XPath(".//text()")(el)).strip())
                list_item_height += list_item_length / inner_width * 70

        return list_height + 10



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


from unittest import mock
import os

from superdesk.tests import TestCase
from superdesk.io.feeding_services import ap
from superdesk.io.feed_parsers import newsml_2_0

PREFIX = 'test_superdesk_'
PROVIDER = {
    "_id": "test_provider",
    "config": {
        "username": "user",
        "password": "password",
        "idList": "123",
        "feed_parser": "newsml2",
        "field_aliases": []
    },
}


class APTestCase(TestCase):

    def setUp(self):
        super().setUp()
        with self.app.app_context():
            vocab = [{}]
            self.app.data.insert('vocabularies', vocab)
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.normpath(os.path.join(dirname, '../fixtures', 'ap.xml'))
        with open(fixture, 'rb') as f:
            self.feed_raw = f.read()

    @mock.patch.object(ap, 'requests')
    @mock.patch.object(ap.APFeedingService, 'get_feed_parser')
    def test_feeding(self, get_feed_parser, requests):
        get_feed_parser.return_value = newsml_2_0.NewsMLTwoFeedParser()
        mock_get = requests.get.return_value
        mock_get.content = self.feed_raw
        provider = PROVIDER.copy()
        service = ap.APFeedingService()
        items = service._update(provider, {})[0]
        self.assertEqual(len(items), 3)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import os
import unittest
import ftplib

from superdesk.publish.transmitters.ftp import FTPPublishService


class FTPPublishTestCase(unittest.TestCase):

    item = {'item_id': 'abc',
            'format': 'NITF',
            'formatted_item': '1234567890'}

    def is_item_loaded(self, url, uploaded_filename):
        config = FTPPublishService().config_from_url(url)
        with ftplib.FTP(config.get('host')) as ftp:
            ftp.login(config.get('username'), config.get('password'))
            ftp.cwd(config.get('path', ''))
            ftp.set_pasv(config.get('passive', False))

            for filename, facts in ftp.mlsd():
                if filename == uploaded_filename:
                    return True
            return False

    def test_it_can_connect(self):
        service = FTPPublishService()

        if 'FTP_URL' not in os.environ:
            return

        config = service.config_from_url(os.environ['FTP_URL'])
        self.item['destination'] = {'config': config}

        self.assertEqual('test', config['path'])
        self.assertEqual('localhost', config['host'])

        service._transmit(self.item, destination={'config': config})
        self.assertTrue(self.is_item_loaded(config, 'abc.ntf'))



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2016 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from superdesk.resource import Resource
from superdesk.metadata.utils import item_url


class SuggestionsResource(Resource):
    """Resource used for live suggestions functionality.
    """

    item_url = item_url
    url = 'suggestions/<{0}:item_id>'.format(item_url)

    item_methods = ['GET']
    privileges = {'GET': 'archive'}
